"""RoomEnv1 environment compatible with gym.

This env uses the RoomDes (room_env/envs/des.py).
This is a more generalized version than RoomEnv0.
"""

import random

import gymnasium as gym

from ..des import RoomDes
from ..utils import seed_everything


class RoomEnv1(gym.Env):
    """The Room environment version 1.
    self.question_sequence
        Every string value is lower-cased to avoid confusion!!!
    """

    metadata = {"render.modes": ["console"]}

    def __init__(
        self,
        des_size: str = "l",
        seed: int = 42,
        question_prob: int = 1.0,
        allow_random_human: bool = False,
        allow_random_question: bool = False,
        check_resources: bool = True,
    ) -> None:
        """

        Args:
            des_size: "xxs", "xs", "s", "m", or "l".
            seed: random seed number
            question_prob: The probability of a question being asked at every observation.
            allow_random_human: whether or not to generate a random human sequence.
            allow_random_question: whether or not to geneate a random question sequence.
            check_resources: whether to check the resources in the DES.

        """
        super().__init__()
        self.seed = seed
        seed_everything(self.seed)
        self.question_prob = question_prob
        self.allow_random_human = allow_random_human
        self.allow_random_question = allow_random_question
        self.check_resources = check_resources

        # Our state / actionspace is quite complex. Here we just make a dummy spaces
        # to bypass the gymnasium sanity check.
        self.observation_space = gym.spaces.Discrete(1)
        self.action_space = gym.spaces.Discrete(1)

        self.des_size = des_size
        self.des = RoomDes(
            des_size=self.des_size,
            check_resources=self.check_resources,
        )
        self.total_maximum_episode_rewards = self.des.until
        assert 0 < self.question_prob <= 1

    def generate_sequences(self) -> None:
        """Generate human and question sequences in advance."""
        if self.allow_random_human:
            self.human_sequence = random.choices(
                list(self.des.humans), k=self.des.until + 1
            )
        else:
            self.human_sequence = (
                self.des.humans * (self.des.until // len(self.des.humans) + 1)
            )[: self.des.until + 1]

        if self.allow_random_question:
            self.question_sequence = [
                random.choice(self.human_sequence[: i + 1])
                for i in range(len(self.human_sequence))
            ]
        else:
            self.question_sequence = [self.human_sequence[0]]
            self.des.run()
            assert (
                len(self.des.states)
                == len(self.des.events) + 1
                == len(self.human_sequence)
            )
            for i in range(len(self.human_sequence) - 1):
                start = max(i + 2 - len(self.des.humans), 0)
                end = i + 2
                humans_observed = self.human_sequence[start:end]

                current_state = self.des.states[end - 1]
                humans_not_changed = []
                for j, human in enumerate(humans_observed):
                    observed_state = self.des.states[start + j]

                    is_changed = False
                    for to_check in ["object", "object_location"]:
                        if (
                            current_state[human][to_check]
                            != observed_state[human][to_check]
                        ):
                            is_changed = True
                    if not is_changed:
                        humans_not_changed.append(human)

                self.question_sequence.append(random.choice(humans_not_changed))

            self.des._initialize()

        effective_question_sequence = []
        for i, question in enumerate(self.question_sequence[:-1]):
            if random.random() < self.question_prob:
                effective_question_sequence.append(question)
            else:
                effective_question_sequence.append(None)
        # The last observation shouldn't have a question
        effective_question_sequence.append(None)
        self.question_sequence = effective_question_sequence

        assert len(self.human_sequence) == len(self.question_sequence)

        self.num_questions = sum(
            [True for question in self.question_sequence if question is not None]
        )

        self.CORRECT = 1
        self.WRONG = -1

    def generate_oqa(
        self, increment_des: bool = False
    ) -> tuple[dict, dict, dict, bool]:
        """Generate an observation, question, and answer.

        Args:
            increment_des: whether or not to take a step in the DES.

        Returns:
            observation: [head, relation, tail, timestamp]
            question: [head, relation, ?, timestamp]
            answer = tail
            is_last: True, if its the last observation in the queue, othewise False

        """
        human_o = self.human_sequence.pop(0)
        human_q = self.question_sequence.pop(0)

        is_last_o = len(self.human_sequence) == 0
        is_last_q = len(self.question_sequence) == 0

        assert is_last_o == is_last_q
        is_last = is_last_o

        if increment_des:
            self.des.step()

        obj_o = self.des.state[human_o]["object"]
        obj_loc_o = self.des.state[human_o]["object_location"]
        observation = [
            f"{human_o}'s {obj_o}",
            "atlocation",
            obj_loc_o,
            self.des.current_time,
        ]
        if human_q is not None:
            obj_q = self.des.state[human_q]["object"]
            obj_loc_q = self.des.state[human_q]["object_location"]

            question = [
                f"{human_q}'s {obj_q}",
                "atlocation",
                "?",
                self.des.current_time,
            ]
            answer = obj_loc_q

        else:
            question = None
            answer = None

        return observation, question, answer, is_last

    def reset(self) -> tuple[dict, dict]:
        """Reset the environment.

        This method somehow can't take arguments. I think it's a bug.


        Returns:
            state, info

        """
        self.des._initialize()
        self.generate_sequences()
        info = {}
        self.obs, self.question, self.answer, self.is_last = self.generate_oqa(
            increment_des=False
        )

        return (self.obs, self.question), info

    def step(self, action: str) -> tuple[tuple, int, bool, bool, dict]:
        """An agent takes an action.

        Args:
            action: An answer to the question.

        Returns:
            state, reward, done, truncated, info

        """
        info = {}
        truncated = False
        if (self.question is None) and (self.answer is None):
            reward = 0
        else:
            if action.lower() == self.answer:
                reward = self.CORRECT
            else:
                reward = self.WRONG

        self.obs, self.question, self.answer, self.is_last = self.generate_oqa(
            increment_des=True
        )

        if self.is_last:
            done = True
        else:
            done = False

        return (self.obs, self.question), reward, done, truncated, info

    def render(self, mode="console") -> None:
        if mode != "console":
            raise NotImplementedError()
        else:
            pass



"""Discrete Event Simlator (RoomDes) for the Room."""

from copy import deepcopy
from pprint import pprint

from .utils import read_json_prod as read_json
from .utils import write_json


def fill_des_resources(des_size: str) -> None:
    """Fill resources

    Args:
        des_size

    """
    des = RoomDes(des_size=des_size, check_resources=False)
    des.run()
    resources = {
        foo: 9999
        for foo in set(
            [bar["object_location"] for foo in des.states for bar in foo.values()]
        )
    }
    des.config["resources"] = deepcopy(resources)
    write_json(des.config, f"./room_env/data/des-config-{des_size}.json")
    resources = []
    des = RoomDes(des_size=des_size, check_resources=True)
    resources.append(deepcopy(des.resources))
    while des.until > 0:
        des.step()
        des.until -= 1
        resources.append(deepcopy(des.resources))

    object_locations = deepcopy(list(des.resources.keys()))
    resources = {
        object_location: 9999
        - min([resource[object_location] for resource in resources])
        for object_location in object_locations
    }

    des.config["resources"] = deepcopy(resources)
    write_json(des.config, f"./room_env/data/des-config-{des_size}.json")
    des = RoomDes(des_size=des_size, check_resources=True)


class RoomDes:
    """RoomDes Class.

    This class is very simple at the moment. When it's initialized, it places N_{humans}
    in the room. They periodically move to other locations. They also periodically
    change the location of their objects they are holding. At the moment,
    everything is deterministic.

    """

    def __init__(self, des_size: str = "l", check_resources: bool = True) -> None:
        """Instantiate the class.

        Args:
            des_size: configuartion for the RoomDes simulation. It should be either size or
                dict. size can be "xxs (extra extra small", "xs (extra small)", "s (small)",
                "m (medium)", or "l (large)".

                {"components": <COMPONENTS>, "resources": <RESOURCES>,
                "last_timestep": <LAST_TIMESTEP>,
                "semantic_knowledge": <SEMANTIC_KNOWLEDGE>, "complexity", <COMPLEXITY>}

                <COMPONENTS> should look like this:

                <RESOURCES> should look like this:

                {'desk': 2, 'A': 10000, 'lap': 10000}

                <LAST_TIMESTEP> is a number where the DES terminates.

                <SEMANTIC_KNOWLEDGE> is a dictionary of semantic knowledge.

                <COMPLEXITY> is defined as num_humans * num_total_objects
                * maximum_num_objects_per_human * maximum_num_locations_per_object

            check_resources: whether to check if the resources are depleted or not.

        """
        if isinstance(des_size, str):
            assert des_size.lower() in [
                "dev",
                "xxs",
                "xs",
                "s",
                "m",
                "l",
            ]
            self.config = read_json(f"./data/des-config-{des_size.lower()}-v1.json")
        else:
            self.config = des_size
        self.check_resources = check_resources
        self._initialize()

    def _initialize(self) -> None:
        """Initialize the simulator."""
        self.components = deepcopy(self.config["components"])
        self.resources = deepcopy(self.config["resources"])
        self.semantic_knowledge = [
            [key, "atlocation", val]
            for key, val in self.config["semantic_knowledge"].items()
        ]

        self.humans = []
        self.objects = []
        self.object_locations = []

        for human, obj_locs in self.components.items():
            self.humans.append(human)

            for obj, loc in obj_locs:
                self.objects.append(obj)
                self.object_locations.append(loc)

        self.humans = sorted(list(set(self.humans)))
        self.objects = sorted(list(set(self.objects)))
        self.object_locations = sorted(list(set(self.object_locations)))

        self.until = deepcopy(self.config["last_timestep"])

        self.states = []
        self.state = {}

        for human in self.components:
            self.state[human] = {}

            (
                self.state[human]["object"],
                self.state[human]["object_location"],
            ) = self.components[human][0]

            if self.check_resources:
                self.resources[self.state[human]["object_location"]] -= 1
            self.state[human]["current_time"] = 0

        self.states.append(deepcopy(self.state))
        if self.check_resources:
            for key, val in self.resources.items():
                assert val >= 0, f"{key}: {val}"

        self.events = []
        self.current_time = 0

    def step(self) -> None:
        """Proceed time by one."""
        previous_state = deepcopy(self.state)
        previous_resources = deepcopy(self.resources)

        self.current_time += 1

        for human in self.state:
            object_location_idx = self.current_time % len(self.components[human])

            self.state[human]["current_time"] = self.current_time

            if self.check_resources:
                self.resources[self.state[human]["object_location"]] += 1

            (
                self.state[human]["object"],
                self.state[human]["object_location"],
            ) = self.components[human][object_location_idx]

            if self.check_resources:
                self.resources[self.state[human]["object_location"]] -= 1

        if self.check_resources:
            for key, val in self.resources.items():
                assert val >= 0, f"{key}: {val}"

        current_state = deepcopy(self.state)
        current_resources = deepcopy(self.resources)
        self.event = self.check_event(
            previous_state, previous_resources, current_state, current_resources
        )
        self.events.append(deepcopy(self.event))
        self.states.append(deepcopy(self.state))

    def check_event(
        self,
        previous_state: dict,
        previous_resources: dict,
        current_state: dict,
        current_resources: dict,
    ) -> dict:
        """Check if any events have occured between the two consecutive states.

        Args:
            previous_state
            previous_resources
            current_state
            current_resources

        Returns:
            event

        """
        assert len(previous_state) == len(current_state)
        assert len(previous_resources) == len(current_resources)

        state_changes = {}
        resource_changes = {}

        humans = list(previous_state)
        for human in humans:
            previous_object = previous_state[human]["object"]
            previous_object_location = previous_state[human]["object_location"]
            previous_time = previous_state[human]["current_time"]

            current_object = current_state[human]["object"]
            current_object_location = current_state[human]["object_location"]
            current_time = current_state[human]["current_time"]

            assert current_time == previous_time + 1

            state_changes[human] = {}

            if previous_object != current_object:
                state_changes[human]["object"] = {
                    "previous": previous_object,
                    "current": current_object,
                }
            if previous_object_location != current_object_location:
                state_changes[human]["object_location"] = {
                    "previous": previous_object_location,
                    "current": current_object_location,
                }

            if len(state_changes[human]) == 0:
                del state_changes[human]
            else:
                state_changes[human]["current_time"] = current_time

        for resource in previous_resources:
            previous_amount = previous_resources[resource]
            current_amount = current_resources[resource]

            if previous_amount != current_amount:
                resource_changes[resource] = current_amount - previous_amount

        return {"state_changes": state_changes, "resource_changes": resource_changes}

    def run(self, debug: bool = False) -> None:
        """Run until the RoomDes terminates."""
        while self.until > 0:
            self.step()
            if debug:
                pprint(self.event)

            self.until -= 1



import unittest

from room_env.utils import argmax, get_duplicate_dicts, list_duplicates_of


class UtilsTest(unittest.TestCase):
    def test_argmax(self):
        self.assertEqual(argmax([6, 1, 2, 3, 4, 5]), 0)

    def test_get_duplicate_dicts(self):
        foo = get_duplicate_dicts({"foo": 1}, [{"foo": 1}, {"bar": 2}, {"foo": 1}])
        self.assertEqual(foo, [{"foo": 1}, {"foo": 1}])

    def test_list_duplicates_of(self):
        foo = list_duplicates_of(
            [{"foo": 1}, {"bar": 2}, {"foo": 2}, {"foo": 1}], {"foo": 1}
        )
        self.assertEqual(foo, [0, 3])



from gymnasium.envs.registration import register

register(
    id="RoomEnv-v0",
    entry_point="room_env.envs:RoomEnv0",
)

register(
    id="RoomEnv-v1",
    entry_point="room_env.envs:RoomEnv1",
)

register(
    id="RoomEnv-v2",
    entry_point="room_env.envs:RoomEnv2",
)

register(
    id="RoomEnv-v3",
    entry_point="room_env.envs:RoomEnv3",
)



"""Utility functions"""

import json
import os
import random
import subprocess
from copy import deepcopy
from typing import Any

import gymnasium as gym
import numpy as np
import torch
import yaml
from rdflib import Graph, URIRef


def sample_max_value_key(
    prob_dict: dict[Any, float], keys_to_exclude: list = None
) -> Any:
    """Sample the key with the maximum value.

    Args:
        prob_dict: dict of probabilities
        keys_to_exclude: keys to exclude

    """
    if keys_to_exclude is not None:
        prob_dict = {k: v for k, v in prob_dict.items() if k not in keys_to_exclude}
    else:
        prob_dict = prob_dict
    max_key = max(prob_dict, key=prob_dict.get)

    return max_key


def seed_everything(seed: int) -> None:
    """Seed every randomness to seed"""
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True


def read_lines(fname: str) -> list:
    """Read lines from a text file.

    There is some path magic going on here. This is to account for both the production
    and development mode. Don't use this function for a general purpose.

    """
    if fname.startswith("/"):
        fullpath = fname
    else:
        fullpath = os.path.join(os.path.dirname(__file__), fname)

    with open(fullpath, "r") as stream:
        names = stream.readlines()
    names = [line.strip() for line in names]

    return names


def read_json(fname: str) -> None:
    """Read json"""
    with open(fname, "r") as stream:
        return json.load(stream)


def write_json(content: dict, fname: str) -> None:
    """Write json"""
    with open(fname, "w") as stream:
        json.dump(content, stream, indent=4, sort_keys=False)


def read_json_prod(fname: str) -> None:
    """Read json.

    There is some path magic going on here. This is to account for both the production
    and development mode. Don't use this function for a general purpose.

    """
    fullpath = os.path.join(os.path.dirname(__file__), fname)

    with open(fullpath, "r") as stream:
        return json.load(stream)


def write_json_prod(content: dict, fname: str) -> None:
    """Write json.

    There is some path magic going on here. This is to account for both the production
    and development mode. Don't use this function for a general purpose.

    """
    fullpath = os.path.join(os.path.dirname(__file__), fname)

    with open(fullpath, "w") as stream:
        json.dump(content, stream, indent=4, sort_keys=False)


def read_yaml(fname: str) -> None:
    """Read yaml.

    There is some path magic going on here. This is to account for both the production
    and development mode. Don't use this function for a general purpose.

    """
    if fname.startswith("/"):
        fullpath = fname
    else:
        fullpath = os.path.join(os.path.dirname(__file__), fname)
    with open(fullpath, "r") as stream:
        return yaml.safe_load(stream)


def write_yaml(content: dict, fname: str) -> None:
    """write yaml."""
    with open(fname, "w") as stream:
        yaml.dump(content, stream, indent=2, sort_keys=False)


def read_data(data_path: str) -> dict:
    """Read train, val, test spilts.

    Args:
        data_path: path to data.

    Returns:
        data: {'train': list of training obs,
            'val': list of val obs,
            'test': list of test obs}

    """
    data = read_json(data_path)

    return data


def argmax(iterable) -> int:
    """argmax"""
    return max(enumerate(iterable), key=lambda x: x[1])[0]


def split_by_possessive(name_entity: str) -> tuple[str, str]:
    """Separate name and entity from the given string.

    Args:
        name_entity: e.g., "tae's laptop"

    Returns:
        name: e.g., tae
        entity: e.g., laptop

    """
    if "'s " in name_entity:
        name, entity = name_entity.split("'s ")
    else:
        name, entity = None, None

    return name, entity


def get_duplicate_dicts(search: dict, target: list) -> list:
    """Find if there are duplicate dicts.

    Args:
        search: dict
        target: target list to look up.

    Returns:
        duplicates: a list of dicts or None

    """
    assert isinstance(search, dict)
    duplicates = []

    for candidate in target:
        assert isinstance(candidate, dict)
        if set(search).issubset(set(candidate)):
            if all([val == candidate[key] for key, val in search.items()]):
                duplicates.append(candidate)

    return duplicates


def list_duplicates_of(seq, item) -> list:
    # https://stackoverflow.com/questions/5419204/index-of-duplicates-items-in-a-python-list
    start_at = -1
    locs = []
    while True:
        try:
            loc = seq.index(item, start_at + 1)
        except ValueError:
            break
        else:
            locs.append(loc)
            start_at = loc
    return locs


def make_des_config(
    commonsense_prob: float,
    num_humans: int,
    num_total_objects: int,
    maximum_num_objects_per_human: int,
    maximum_num_locations_per_object: int,
    maxiumum_days_period: int,
    des_size: str,
    last_timestep: int = 128,
) -> dict:
    """Make a des config.

    Args:
        commonsense_prob: commonsense probability
        num_humans: number of humans
        num_total_objects: number of total objects
        maximum_num_objects_per_human: maximum number of objects per human
        maximum_num_locations_per_object: maximum number of locations per object
        maxiumum_days_period: maxiumum number of days period
        des_size: The size of DES (i.e., "xxs", "xs", "s", "m", "l", "dev")
        last_timestep: last time step where the DES terminates.

    Returns:
        des config

    """
    des_config = {
        "human_names_path": "./room_env/data/human-names",
        "last_timestep": last_timestep,
        "maxiumum_days_period": maxiumum_days_period,
        "save_path": f"./room_env/data/des-config-{des_size}.json",
        "seed": 42,
        "semantic_knowledge_path": "./room_env/data/semantic-knowledge.json",
    }

    des_config["commonsense_prob"] = commonsense_prob
    des_config["num_humans"] = num_humans
    des_config["num_total_objects"] = num_total_objects
    des_config["maximum_num_objects_per_human"] = maximum_num_objects_per_human
    des_config["maximum_num_locations_per_object"] = maximum_num_locations_per_object

    return des_config


def get_des_variables(des_size: str = "l") -> tuple[int, int, int]:
    """Get the des variables.

    Args:
        des_size: The size of DES (i.e., "xxs", "xs", "s", "m", "l", "dev")

    Returns:
        capacity, num_humans, num_total_objects

    """
    if des_size == "dev":
        capacity = 16

    elif des_size == "xxs":
        capacity = 2

    elif des_size == "xs":
        capacity = 4

    elif des_size == "s":
        capacity = 8

    elif des_size == "m":
        capacity = 16

    elif des_size == "l":
        capacity = 32

    else:
        raise ValueError

    num_humans = capacity * 2
    num_total_objects = capacity // 2

    return capacity, num_humans, num_total_objects


def run_des_seeds(
    seeds: list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    capacity: int = 32,
    des_size: str = "l",
    allow_random_human: bool = True,
    allow_random_question: bool = True,
    question_prob: float = 0.1,
) -> dict:
    """Run the RoomEnv-v1 with multiple different seeds.

    Args:
        seeds:
        capacity:
        des_size:
        allow_random_human:
        allow_random_question:
        question_prob:

    Returns:
        results

    """
    results = {}
    how_to_forget = ["episodic", "semantic", "random", "pre_sem"]

    for forget_short in how_to_forget:
        if forget_short == "random":
            pretrain_semantic = False
            capacity_ = {
                "episodic": capacity // 2,
                "semantic": capacity // 2,
                "short": 1,
            }
        elif forget_short == "episodic":
            pretrain_semantic = False
            capacity_ = {"episodic": capacity, "semantic": 0, "short": 1}
        elif forget_short == "semantic":
            pretrain_semantic = False
            capacity_ = {"episodic": 0, "semantic": capacity, "short": 1}
        elif forget_short == "pre_sem":
            pretrain_semantic = True
            capacity_ = {
                "episodic": capacity // 2,
                "semantic": capacity // 2,
                "short": 1,
            }
        else:
            raise ValueError

        results_ = []
        for seed in seeds:
            env = gym.make(
                "room_env:RoomEnv-v1",
                des_size=des_size,
                seed=seed,
                policies={
                    "memory_management": "rl",
                    "answer_question": "episodic_semantic",
                    "encoding": "argmax",
                },
                capacity=capacity_,
                question_prob=question_prob,
                observation_params="perfect",
                allow_random_human=allow_random_human,
                allow_random_question=allow_random_question,
                pretrain_semantic=pretrain_semantic,
                check_resources=False,
                varying_rewards=False,
            )
            state, info = env.reset()
            rewards = 0
            while True:
                if forget_short == "random":
                    action = random.choice([0, 1, 2])
                elif forget_short == "episodic":
                    action = 0
                elif forget_short == "semantic":
                    action = 1
                elif forget_short == "pre_sem":
                    action = 0
                else:
                    raise ValueError
                state, reward, done, truncated, info = env.step(action)
                rewards += reward
                if done:
                    break
            results_.append(rewards)

        results[forget_short] = np.mean(results_)

    return results


def run_all_des_configs(
    des_size: str,
    capacity: int,
    maximum_num_objects_per_human: int,
    maximum_num_locations_per_object: int,
    maxiumum_days_period: int,
    commonsense_prob: float,
    num_humans: int,
    num_total_objects: int,
    seeds: list,
    allow_random_human: bool,
    allow_random_question: bool,
    last_timestep: int,
    question_prob: float,
) -> dict:
    """Run the RoomEnv-v1 with different des configs, with multiple different seeds.

    Args:
        des_size: The size of DES (i.e., "xxs", "xs", "s", "m", "l", "dev")
        capacity: int,
        maximum_num_objects_per_human: maximum number of objects per human
        maximum_num_locations_per_object: maximum number of locations per object
        maxiumum_days_period: maxiumum number of days period
        commonsense_prob: commonsense probability
        num_humans: number of humans
        num_total_objects: number of total objects
        seeds: list,
        allow_random_human: bool,
        allow_random_question: bool,
        last_timestep: int,
        question_prob: float,

    Returns:
        results

    """
    des_config = make_des_config(
        commonsense_prob=commonsense_prob,
        num_humans=num_humans,
        num_total_objects=num_total_objects,
        maximum_num_objects_per_human=maximum_num_objects_per_human,
        maximum_num_locations_per_object=maximum_num_locations_per_object,
        maxiumum_days_period=maxiumum_days_period,
        des_size=des_size,
        last_timestep=last_timestep,
    )

    complexity = (
        num_humans
        * num_total_objects
        * maximum_num_objects_per_human
        * maximum_num_locations_per_object
        * maxiumum_days_period
    )

    with open("create_des_config.yaml", "w") as stream:
        yaml.safe_dump(des_config, stream, indent=2)

    sub_out = subprocess.call(
        ["python", "create_des_config.py"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    if sub_out == 1:
        return None

    rewards = run_des_seeds(
        seeds=seeds,
        capacity=capacity,
        des_size=des_size,
        allow_random_human=allow_random_human,
        allow_random_question=allow_random_question,
        question_prob=question_prob,
    )
    results = {
        "mean_rewards_diff": rewards["pre_sem"]
        - rewards["random"] / 3
        - rewards["semantic"] / 3
        - rewards["episodic"] / 3,
        "mean_rewards_episodic": rewards["episodic"],
        "mean_rewards_semantic": rewards["semantic"],
        "mean_rewards_random": rewards["random"],
        "mean_rewards_pre_sem": rewards["pre_sem"],
        "complexity": complexity,
        "commonsense_prob": commonsense_prob,
        "maximum_num_locations_per_object": maximum_num_locations_per_object,
        "maximum_num_objects_per_human": maximum_num_objects_per_human,
        "num_humans": num_humans,
        "num_total_objects": num_total_objects,
        "maxiumum_days_period": maxiumum_days_period,
        "allow_random_human": allow_random_human,
        "allow_random_question": allow_random_question,
        "question_prob": question_prob,
    }
    return results


def get_handcrafted(
    env: str = "room_env:RoomEnv-v1",
    des_size: str = "l",
    seeds: list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    question_prob: float = 0.1,
    observation_params: str = "perfect",
    policies: dict = {
        "memory_management": "rl",
        "answer_question": "episodic_semantic",
        "encoding": "argmax",
    },
    capacities: list = [2, 4, 8, 16, 32, 64],
    allow_random_human: bool = False,
    allow_random_question: bool = True,
    varying_rewards: bool = False,
    check_resources: bool = True,
) -> None:
    """Get the env results with handcrafted policies.

    At the moment only {"memory_management": "rl"} is supported.

    Args:
        env: str = "RoomEnv-v1",
        des_size: str = "l",
        seeds: list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        question_prob: float = 0.1,
        policies: dict = {
            "memory_management": "rl",
            "answer_question": "episodic_semantic",
            "encoding": "argmax",
        },
        capacities: list = [2, 4, 8, 16, 32, 64],
        allow_random_human: whether to allow random humans to be observed.
        allow_random_question: whether to allow random questions to be asked.
        varying_rewards: If true, then the rewards are scaled in every episode so that
                total_maximum_episode_rewards is 128.

    Returns:
        handcrafted_results

    """
    how_to_forget = ["episodic", "semantic", "random", "pre_sem"]
    env_ = env
    handcrafted_results = {}

    for capacity in capacities:
        handcrafted_results[capacity] = {}
        for forget_short in how_to_forget:
            if forget_short == "random":
                pretrain_semantic = False
                capacity_ = {
                    "episodic": capacity // 2,
                    "semantic": capacity // 2,
                    "short": 1,
                }
            elif forget_short == "episodic":
                pretrain_semantic = False
                capacity_ = {"episodic": capacity, "semantic": 0, "short": 1}
            elif forget_short == "semantic":
                pretrain_semantic = False
                capacity_ = {"episodic": 0, "semantic": capacity, "short": 1}
            elif forget_short == "pre_sem":
                pretrain_semantic = True
                capacity_ = {
                    "episodic": capacity // 2,
                    "semantic": capacity // 2,
                    "short": 1,
                }
            else:
                raise ValueError

            results = []
            for seed in seeds:
                env = gym.make(
                    env_,
                    des_size=des_size,
                    seed=seed,
                    policies=policies,
                    capacity=capacity_,
                    question_prob=question_prob,
                    observation_params=observation_params,
                    allow_random_human=allow_random_human,
                    allow_random_question=allow_random_question,
                    pretrain_semantic=pretrain_semantic,
                    check_resources=check_resources,
                    varying_rewards=varying_rewards,
                )
                state, info = env.reset()
                rewards = 0
                while True:
                    if forget_short == "random":
                        action = random.choice([0, 1, 2])
                    elif forget_short == "episodic":
                        action = 0
                    elif forget_short == "semantic":
                        action = 1
                    elif forget_short == "pre_sem":
                        action = 0
                    else:
                        raise ValueError
                    state, reward, done, truncated, info = env.step(action)
                    rewards += reward
                    if done:
                        break
                results.append(rewards)

            mean_ = np.mean(results).round(3).item()
            std_ = np.std(results).round(3).item()
            handcrafted_results[capacity][forget_short] = {"mean": mean_, "std": std_}

    return handcrafted_results


def find_connected_nodes(graph_):
    graph = deepcopy(graph_)

    def dfs(node, row, col):
        if (
            row < 0
            or col < 0
            or row >= len(graph)
            or col >= len(graph[0])
            or graph[row][col] == 0
        ):
            return

        connected_nodes.append((row, col))
        graph[row][col] = 0  # Mark the node as visited

        # Check the neighbors
        dfs(node, row - 1, col)  # Up
        dfs(node, row + 1, col)  # Down
        dfs(node, row, col - 1)  # Left
        dfs(node, row, col + 1)  # Right

    connected_components = []
    for row in range(len(graph)):
        for col in range(len(graph[row])):
            if graph[row][col] == 1:
                connected_nodes = []
                dfs(1, row, col)
                if connected_nodes:
                    connected_components.append(connected_nodes)

    return connected_components


def is_running_notebook() -> bool:
    """See if the code is running in a notebook or not."""
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def rdf_to_dict(g: Graph) -> dict:
    """Convert an rdflib Graph to a Python dictionary representation."""
    graph_dict = {}
    for s, p, o in g:
        s_key = str(s)  # Convert subject to string
        p_key = str(p)  # Convert predicate to string
        o_value = str(o)  # Convert object to string

        # Initialize subject entry if not present
        if s_key not in graph_dict:
            graph_dict[s_key] = {}

        # If predicate exists, store as a list (for multiple values)
        if p_key in graph_dict[s_key]:
            if isinstance(graph_dict[s_key][p_key], list):
                graph_dict[s_key][p_key].append(o_value)
            else:
                graph_dict[s_key][p_key] = [graph_dict[s_key][p_key], o_value]
        else:
            graph_dict[s_key][p_key] = o_value

    return graph_dict


def dict_to_rdf(graph_dict: dict) -> Graph:
    """Convert a dictionary representation to an RDFLib Graph object."""
    g = Graph()
    for s, preds in graph_dict.items():
        for p, objs in preds.items():
            if isinstance(objs, list):
                for o in objs:
                    g.add((URIRef(s), URIRef(p), URIRef(o)))
            else:
                g.add((URIRef(s), URIRef(p), URIRef(objs)))
    return g


def list_to_rdf(triples: list[list[str]]) -> Graph:
    """Convert a list of triples (list of lists) into an RDFLib Graph object."""
    g = Graph()
    for s, p, o in triples:
        s_ref, p_ref, o_ref = URIRef(s), URIRef(p), URIRef(o)
        g.add((s_ref, p_ref, o_ref))
    return g


def rdf_to_list(g: Graph) -> list[list[str]]:
    """Convert an rdflib Graph to a list of triples (list of lists)."""
    triples = []
    for s, p, o in g:
        triples.append([str(s), str(p), str(o)])
    return triples


def list_to_dict(triples: list[list[str]]) -> dict:
    """Convert a list of triples (list of lists) into a dictionary representation."""
    graph_dict = {}
    for s, p, o in triples:
        if s not in graph_dict:
            graph_dict[s] = {}
        if p in graph_dict[s]:
            if isinstance(graph_dict[s][p], list):
                graph_dict[s][p].append(o)
            else:
                graph_dict[s][p] = [graph_dict[s][p], o]
        else:
            graph_dict[s][p] = o
    return graph_dict


def dict_to_list(graph_dict: dict) -> list[list[str]]:
    """
    Convert a dictionary representation to a list of triples (list of lists).
    """
    triples = []
    for s, preds in graph_dict.items():
        for p, objs in preds.items():
            if isinstance(objs, list):
                for o in objs:
                    triples.append([s, p, o])
            else:
                triples.append([s, p, objs])
    return triples



"""RoomEnv0 environment compatible with gym."""

import random
from itertools import cycle

import gymnasium as gym

from ..utils import read_json_prod as read_json
from ..utils import read_lines, split_by_possessive

CORRECT = 1
WRONG = 0


class RoomEnv0(gym.Env):
    """The Room environment version 0.

    In this big room, N_{agents} move around and observe N_{humans} placing objects.
    Every time the agents move around (i.e., take a step), they observe one human_{i}
    placing an object somewhere. Each agent can only observe one human at a time.
    Every time the agents takes a step, the environment also asks them where an object
    is. +1 reward is given when it gets right and 0 when it gets wrong.

    This environment is challenging in two ways:

    (1) An agent can't observe the entire room. It can only observe one human at a time.
    This means that the environment is only partially observable. This constraint means
    that it's more beneficial if more than one agent can collaborate. This constraint
    also means that the agents should have a memory system to remember the past
    observations.

    (2) The room environment changes every time. The humans can switch their locations,
    change their objects, and place them at different locations. These changes are
    not completely random. A decent portion of them come from commmon-sense knowledge.
    This means that an agent with both episodic and semantic memory systems will perform
    better than an agent with only one memory system.

    Every string value is lower-cased to avoid confusion!!!

    """

    metadata = {"render.modes": ["console"]}

    def __init__(
        self,
        room_size: str = "small",
        probs: dict = {
            "commonsense": 0.7,
            "new_location": 0.1,
            "new_object": 0.1,
            "switch_person": 0.5,
        },
        limits: dict = {
            "heads": None,
            "tails": None,
            "names": None,
            "allow_spaces": True,
        },
        max_step: int = 1000,
        disjoint_entities: bool = True,
        num_agents: int = 1,
    ) -> None:
        """Initialize the environment.

        Args:
            room_size: small or big
            probs: the probabilities that govern the room environment changes.
            limits: Limitation on the triples.
            max_step: maximum step an agent can take. The environment will terminate when
                the number reaches this value.
            disjoint_entities: Assure that the heads and the tails don't overlap.
            num_agents: number of agents in the room.

        """
        super().__init__()
        self.room_size = room_size
        if self.room_size.lower() == "small":
            semantic_knowledge_path = "./data/semantic-knowledge-small.json"
            names_path = "./data/top-human-names-small"

        elif self.room_size.lower() in ["large", "big"]:
            semantic_knowledge_path = "./data/semantic-knowledge.json"
            names_path = "./data/top-human-names"

        self.limits = limits

        (
            self.semantic_knowledge,
            self.heads,
            self.relations,
            self.tails,
        ) = self.load_semantic_knowledge(
            semantic_knowledge_path,
            limit_heads=self.limits["heads"],
            limit_tails=self.limits["tails"],
            allow_spaces=self.limits["allow_spaces"],
            disjoint_entities=disjoint_entities,
        )

        assert len(self.relations) == 1, "At the moment there is only one relation."

        self.names = self.read_names(
            names_path, self.limits["names"], self.limits["allow_spaces"]
        )

        assert len(self.names) <= len(self.heads)

        if disjoint_entities:
            lhs = len(set(self.relations + self.names + self.heads + self.tails))
            rhs = (
                len(self.relations)
                + len(self.names)
                + len(self.heads)
                + len(self.tails)
            )

            assert lhs == rhs

        self.probs = probs
        self.max_step = max_step
        self.num_agents = num_agents
        # Our state / action space is quite complex. Here we just make a dummy
        # observation space to bypass the sanity check.
        self.observation_space = gym.spaces.Discrete(1)
        self.action_space = gym.spaces.Discrete(1)

    def reset(self) -> tuple:
        """Reset the environment.

        This will place N_{humans} humans in the room. Each human only has one object,
        which will be placed by the human in a random location.

        """
        self.current_time = 0
        random.shuffle(self.names)
        random.shuffle(self.heads)
        self.room = []

        for name, head in zip(self.names, self.heads):
            relation = self.relations[0]  # At the moment there is only one relation.
            tail = self.generate_tail(head, relation)
            if tail is not None:
                self.room.append([f"{name}'s {head}", relation, f"{tail}"])

        navigate = [[i for i in range(len(self.room))] for _ in range(self.num_agents)]
        for navigate_ in navigate:
            random.shuffle(navigate_)

        self.navigate = [cycle(navigate_) for navigate_ in navigate]

        observations = self.generate_observations()
        question, self.answer = self.generate_qa()

        info = {}

        return (observations, question), info

    def generate_observations(self) -> list:
        """Generate a random obseration.

        Returns:
            observations: e.g., ["tae's laptop, "atlocation", "tae's desk", 10]
                The last element in the list accounts for the timestamp.

        """
        observations = {
            i: [*self.room[next(navigate_)], self.current_time]
            for i, navigate_ in enumerate(self.navigate)
        }

        return observations

    def generate_qa(self) -> tuple[list, str]:
        """Generate a question and the answer.

        Returns:
            question: e.g., ["tae's laptop", "atlocation", "?", 10]
            answer: e.g., desk

        """
        random_choice = random.choice(self.room)
        question = random_choice[:2] + ["?"] + [self.current_time]
        answer = random_choice[2]

        return question, answer

    def generate_tail(self, head: str, relation: str) -> str:
        """This simulates humans placing their objects in their desired locations.

        Note that "head" shouldn't include a human name.

        """
        relevant_tails = [
            triple[2]
            for triple in self.semantic_knowledge
            if triple[0] == head and triple[1] == relation
        ]
        assert len(relevant_tails) == 1

        if random.random() < self.probs["commonsense"]:
            tail = relevant_tails[0]
        else:
            while True:
                tail = random.choice(self.tails)

                if tail not in relevant_tails:
                    break

        return tail

    def renew(self) -> None:
        """Renew the room.

        This is done every time when the agent takes a step. This is doen to simulate
        that the room changes. People move. They place their objects in different
        locations. They also change their objects. All of these are done in a random
        manner. The randomness can be adjusted from the argument `probs`.

        With the chance of probs["new_location"], an object will be placed at a new
        location.

        When the object is placed at a new locaiton, with the chance of
        probs["commonsense"], an object will be placed at a commonsense-knowledge spot.

        With the chance of probs["new_object"], the person with the object will have
        a new random object.

        With the chance of probs["switch_person"], two persons switch their spots.

        """
        room = []
        for head, relation, tail in self.room:
            name1, head = split_by_possessive(head)

            if random.random() < self.probs["new_object"]:
                while True:
                    new_head = random.choice(self.heads)
                    if new_head != head:
                        head = new_head
                        tail = self.generate_tail(head, relation)
                        break
            else:
                if random.random() < self.probs["new_location"]:
                    while True:
                        new_tail = self.generate_tail(head, relation)
                        if new_tail != tail:
                            tail = new_tail
                            break

            room.append(
                [f"{name1}'s {head}", relation, f"{tail}"],
            )

        if random.random() < self.probs["switch_person"]:
            i, j = random.sample(
                range(
                    0,
                    len(
                        self.room,
                    ),
                ),
                2,
            )
            room[i], room[j] = room[j], room[i]

        self.room = room

    def step(self, action: str) -> tuple[tuple, int, bool, bool, dict]:
        """An agent takes an action.

        Args:
            action: This is the agent's answer to the previous question.

        Returns:
            observations: e.g., ["tae's laptop, "atlocation", "tae's desk", 10]
            question: e.g., ["tae's laptop", "atlocation", "?", 10]
            answer: e.g., desk
            reward: +1 for correct answer, 0 for wrong answer.
            done: True if the current_time reaches the max_step.
            truncated: False

        """
        if str(action).lower() == self.answer.lower():
            reward = CORRECT

        else:
            reward = WRONG

        self.current_time += 1

        # Things will change in the room!
        self.renew()

        observations = self.generate_observations()
        question, self.answer = self.generate_qa()

        info = {}

        if self.current_time >= self.max_step:
            done = True
        else:
            done = False

        truncated = False

        return (observations, question), reward, done, truncated, info

    def render(self, mode="console") -> None:
        if mode != "console":
            raise NotImplementedError()
        else:
            pass

    def close(self):
        pass

    @staticmethod
    def read_names(
        path: str = "./data/top-human-names",
        limit_names: int = None,
        allow_spaces: bool = False,
    ) -> list:
        """Read 20 most common names.

        Args:
            path: The path to the top 20 human name list.
            limit_names: Limit the number of names
            allow_spaces: Whether to include words that have spaces
                (e.g., corner of two streets)

        Returns:
            names: human names (e.g., james)

        """
        names = read_lines(path)

        if not allow_spaces:
            names = [name for name in names if len(name.split("_")) == 1]

        if limit_names:
            names = sorted(names, key=len)
            names = names[:limit_names]

        return names

    @staticmethod
    def load_semantic_knowledge(
        path: str,
        limit_heads: int = None,
        limit_tails: int = None,
        allow_spaces: bool = False,
        disjoint_entities: bool = True,
    ) -> tuple[list, list, list, list]:
        """Load saved semantic knowledge.

        Args:
            path: the path to the pretrained semantic memory.
            limit_heads: Limit the number of heads (e.g., 10)
            limit_tails: Limit the number of tails per heads (e.g., 1)
            allow_spaces: Whether to include words that have spaces
                (e.g., corner of two streets)
            disjoint_entities: Whether to force that there are no common elements between
                entities.

        Returns:
            semantic_knowledge
            heads
            relations
            tails

        """
        semantic_knowledge = read_json(path)

        heads = sorted(list(set(semantic_knowledge.keys())))
        if disjoint_entities:
            semantic_knowledge = {
                key: {
                    key_: [tail for tail in val_ if tail["tail"] not in heads]
                    for key_, val_ in val.items()
                }
                for key, val in semantic_knowledge.items()
            }

        semantic_knowledge = {
            key: {
                key_: val_
                for key_, val_ in val.items()
                if len([tail for tail in val_ if tail["tail"]]) > 0
            }
            for key, val in semantic_knowledge.items()
        }
        semantic_knowledge = {
            key: val for key, val in semantic_knowledge.items() if len(val) > 0
        }

        # sort the semantic knowledge by its highest weight to be sure.
        semantic_knowledge = {
            key: {
                key_: sorted(val_, key=lambda x: -x["weight"])
                for key_, val_ in val.items()
            }
            for key, val in semantic_knowledge.items()
        }

        if not allow_spaces:
            semantic_knowledge = {
                key: {
                    key_: [v for v in val_ if len(v["tail"].split("_")) == 1]
                    for key_, val_ in val.items()
                }
                for key, val in semantic_knowledge.items()
                if (len(key.split("_"))) == 1
            }

        if limit_heads:
            semantic_knowledge = {
                key: val
                for idx, (key, val) in enumerate(semantic_knowledge.items())
                if idx < limit_heads
            }

        if limit_tails:
            semantic_knowledge = {
                key: {key_: val_[:limit_tails] for key_, val_ in val.items()}
                for key, val in semantic_knowledge.items()
            }

        heads = sorted(list(set(semantic_knowledge.keys())))
        tails = sorted(
            list(
                set(
                    [
                        foo["tail"]
                        for key, val in semantic_knowledge.items()
                        for key_, val_ in val.items()
                        for foo in val_
                    ]
                )
            )
        )
        relations = sorted(
            list(
                set(
                    [
                        key_
                        for key, val in semantic_knowledge.items()
                        for key_, val_ in val.items()
                    ]
                )
            )
        )

        semantic_knowledge_list = []
        for key, val in semantic_knowledge.items():
            for k, v in val.items():
                semantic_knowledge_list.append([key, k, v[0]["tail"]])

        return semantic_knowledge_list, heads, relations, tails



"""Room Environment v3 - Simplified deterministic version."""

import os
import random

import gymnasium as gym
import matplotlib.pyplot as plt
import networkx as nx
from rdflib import Graph, URIRef

from ..utils import is_running_notebook, rdf_to_list
from ..utils import read_json_prod as read_json


class RoomEnv3(gym.Env):
    """Simplified Room Environment v3 with deterministic object movement."""

    def __init__(
        self,
        terminates_at: int = 99,
        room_size: str = "small",
    ) -> None:
        """Initialize Room Environment v3.

        Args:
            terminates_at: When episode terminates (can be any value)
            room_size: Room configuration size
        """
        super().__init__()

        self.is_notebook = is_running_notebook()

        # Load configuration
        config = read_json(f"data/room-config-{room_size}-v3.json")

        self.grid_length = config["grid_length"]
        self.room_names = config["room_names"]
        self.room_positions = config["room_positions"]
        self.base_room_connections = config["room_connections"]
        self.static_names = config["static_names"]
        self.moving_names = config["moving_names"]
        self.initial_static_locations = config["static_locations"]
        self.initial_moving_locations = config["moving_locations"]
        self.movement_preferences = config["movement_preferences"]
        self.initial_agent_location = config["agent_location"]
        self.selected_walls = config["selected_walls"]

        # Convert wall configs from string keys back to tuples
        self.wall_configs = {}
        for wall_key, pattern in config["wall_configs"].items():
            # Convert string key back to tuple
            parts = wall_key.split("|")
            wall_tuple = (parts[0], parts[1], parts[2])
            self.wall_configs[wall_tuple] = pattern

        self.question_objects = config["question_objects"]  # Always 100 questions

        # Environment parameters
        self.terminates_at = terminates_at

        # Dummy gym spaces
        self.observation_space = gym.spaces.Discrete(1)
        self.action_space = gym.spaces.Discrete(1)

        self.entities = (
            self.room_names + self.static_names + self.moving_names + ["agent", "wall"]
        )
        self.relations = ["north", "east", "south", "west", "at_location"]
        self.total_maximum_episode_rewards = self.terminates_at + 1

    def reset(self):
        """Reset environment to initial state."""
        self.current_step = 0

        # Reset object locations
        self.static_locations = self.initial_static_locations.copy()
        self.moving_locations = self.initial_moving_locations.copy()
        self.agent_location = self.initial_agent_location

        # Set initial wall layout
        self._update_wall_layout()

        # Get initial observations
        observations = self._get_observations()
        info = {}

        return observations, info

    def step(self, actions):
        """Take environment step.

        Args:
            actions: Tuple of (question_answer, movement_action)

        Returns:
            observations, reward, done, truncated, info
        """
        question_answer, movement_action = actions

        # Calculate reward for current question
        reward = self._calculate_reward(question_answer)

        # Move objects deterministically
        self._move_objects()

        # Move agent
        self._move_agent(movement_action)

        # Update step counter
        self.current_step += 1

        # Update wall layout based on patterns
        self._update_wall_layout()

        # Check if done
        done = self.current_step > self.terminates_at
        truncated = False

        # Get next observations
        observations = self._get_observations()
        info = {}

        return observations, reward, done, truncated, info

    def _update_wall_layout(self):
        """Update room connections based on periodic wall patterns."""
        # Start with base connections
        self.room_connections = {}
        for room_name in self.room_names:
            self.room_connections[room_name] = self.base_room_connections[
                room_name
            ].copy()

        # Apply inner walls based on their patterns
        for wall, pattern in self.wall_configs.items():
            # Get pattern state at current step using actual pattern length
            pattern_length = len(pattern)
            pattern_index = self.current_step % pattern_length
            is_wall_active = pattern[pattern_index] == 1

            if is_wall_active:
                room1, room2, wall_type = wall
                if wall_type == "horizontal":
                    self.room_connections[room1]["south"] = "wall"
                    self.room_connections[room2]["north"] = "wall"
                else:  # vertical
                    self.room_connections[room1]["east"] = "wall"
                    self.room_connections[room2]["west"] = "wall"

    def _move_objects(self):
        """Move all moving objects deterministically."""
        for obj_name in self.moving_names:
            current_location = self.moving_locations[obj_name]
            preferences = self.movement_preferences[obj_name]

            # Try each preferred direction in order
            for direction in preferences:
                next_location = self.room_connections[current_location][direction]
                if next_location != "wall":
                    self.moving_locations[obj_name] = next_location
                    break

    def _move_agent(self, action):
        """Move agent based on action."""
        if action in ["north", "east", "south", "west"]:
            next_location = self.room_connections[self.agent_location][action]
            if next_location != "wall":
                self.agent_location = next_location
        # "stay" action or invalid action keeps agent in place

    def _get_observations(self):
        """Get current observations."""
        # Room layout from agent's perspective
        agent_room_obs = []
        for direction in ["north", "east", "south", "west"]:
            connected = self.room_connections[self.agent_location][direction]
            agent_room_obs.append([self.agent_location, direction, connected])

        # Objects in agent's room
        for obj_name in self.static_names + self.moving_names:
            obj_location = self.static_locations.get(
                obj_name
            ) or self.moving_locations.get(obj_name)
            if obj_location == self.agent_location:
                agent_room_obs.append([obj_name, "at_location", obj_location])

        # Agent location
        agent_room_obs.append(["agent", "at_location", self.agent_location])

        random.shuffle(agent_room_obs)

        # Generate exactly one question for this step (not a list)
        question_index = self.current_step % 100  # Cycle through 100 questions
        obj_name = self.question_objects[question_index]
        question = [obj_name, "at_location", "?"]

        return {"room": agent_room_obs, "question": question}

    def _calculate_reward(self, answer):
        """Calculate reward for question answer."""
        question_index = self.current_step % 100
        obj_name = self.question_objects[question_index]

        # Get correct location
        if obj_name in self.static_names:
            correct_location = self.static_locations[obj_name]
        else:
            correct_location = self.moving_locations[obj_name]

        # Check answer and return single reward
        return 1 if answer == correct_location else 0

    def _find_objects_in_room(self, room_name: str) -> list:
        """Find all objects currently in a specified room, categorized by type."""
        objects_in_room = []

        # Check static objects
        for obj_name in self.static_names:
            if self.static_locations[obj_name] == room_name:
                objects_in_room.append(("static", obj_name))

        # Check moving objects
        for obj_name in self.moving_names:
            if self.moving_locations[obj_name] == room_name:
                objects_in_room.append(("moving", obj_name))

        # Check agent
        if self.agent_location == room_name:
            objects_in_room.append(("agent", "agent"))

        return objects_in_room

    def render(
        self,
        render_mode: str = "grid",
        figsize: tuple[int, int] = (12, 12),
        cell_text_size: int = 12,
        save_fig_dir: str = "./DEBUG/",
        image_format: str = "png",
        graph_layout: str = "spring",
    ) -> None:
        """Render the current state of the environment.

        Args:
            render_mode: How to render ('console', 'grid', or 'graph')
            figsize: Size of the figure
            cell_text_size: Text size in cells
            save_fig_dir: Directory to save figures
            image_format: Format to save images in
            graph_layout: Layout for graph rendering ('spring', 'circular', 'kamada_kawai')
        """
        if render_mode == "console":
            print(f"Step {self.current_step}:")
            print(f"Agent location: {self.agent_location}")
            print(f"Static objects: {self.static_locations}")
            print(f"Moving objects: {self.moving_locations}")

        elif render_mode == "graph":
            self._render_graph(figsize, save_fig_dir, image_format, graph_layout)

        elif render_mode == "grid":
            plt.figure(figsize=figsize)

            # Define colors for different object types
            colors = {
                "room": "#FFE4B5",  # Moccasin (Soft Gold)
                "moving": "#90EE90",  # Light Green
                "static": "#87CEFA",  # Light Blue
                "agent": "#D8BFD8",  # Thistle (Soft Purple)
            }

            # Draw the grid
            for i in range(self.grid_length):
                for j in range(self.grid_length):
                    room_idx = i * self.grid_length + j
                    room_name = self.room_names[room_idx]

                    # Draw room cell with room color background
                    rect = plt.Rectangle(
                        (j, self.grid_length - 1 - i),
                        1,
                        1,
                        facecolor=colors["room"],
                        edgecolor="black",
                        linewidth=1,
                    )
                    plt.gca().add_patch(rect)

                    # Get objects in this room
                    objects_in_room = self._find_objects_in_room(room_name)

                    # Group objects by type for display
                    static_objs = [
                        obj for obj_type, obj in objects_in_room if obj_type == "static"
                    ]
                    moving_objs = [
                        obj for obj_type, obj in objects_in_room if obj_type == "moving"
                    ]
                    agent_objs = [
                        obj for obj_type, obj in objects_in_room if obj_type == "agent"
                    ]

                    y_offset = 0.85  # Start position for room name
                    line_height = 0.12  # Space between lines

                    # Display room name at top in black (no background box)
                    plt.text(
                        j + 0.5,
                        self.grid_length - 1 - i + y_offset,
                        room_name,
                        ha="center",
                        va="center",
                        fontsize=cell_text_size,
                        color="black",
                        weight="bold",
                    )
                    y_offset -= line_height * 1.5

                    # Display each static object separately with blue background
                    for obj in static_objs:
                        if y_offset > 0.1:  # Only display if there's space
                            plt.text(
                                j + 0.5,
                                self.grid_length - 1 - i + y_offset,
                                obj,
                                ha="center",
                                va="center",
                                fontsize=cell_text_size - 2,
                                color="black",
                                weight="bold",
                                bbox=dict(
                                    boxstyle="round,pad=0.15",
                                    facecolor=colors["static"],
                                    alpha=0.8,
                                ),
                            )
                            y_offset -= line_height

                    # Display each moving object separately with green background
                    for obj in moving_objs:
                        if y_offset > 0.1:  # Only display if there's space
                            plt.text(
                                j + 0.5,
                                self.grid_length - 1 - i + y_offset,
                                obj,
                                ha="center",
                                va="center",
                                fontsize=cell_text_size - 2,
                                color="black",
                                weight="bold",
                                bbox=dict(
                                    boxstyle="round,pad=0.15",
                                    facecolor=colors["moving"],
                                    alpha=0.8,
                                ),
                            )
                            y_offset -= line_height

                    # Display agent with purple background
                    for obj in agent_objs:
                        if y_offset > 0.1:  # Only display if there's space
                            plt.text(
                                j + 0.5,
                                self.grid_length - 1 - i + y_offset,
                                obj,
                                ha="center",
                                va="center",
                                fontsize=cell_text_size - 1,
                                color="black",
                                weight="bold",
                                bbox=dict(
                                    boxstyle="round,pad=0.15",
                                    facecolor=colors["agent"],
                                    alpha=0.8,
                                ),
                            )

            # Draw walls with thicker lines
            self._draw_walls()

            plt.gca().set_xlim(0, self.grid_length)
            plt.gca().set_ylim(0, self.grid_length)
            plt.gca().set_aspect("equal")

            # Remove axis ticks and labels
            plt.gca().set_xticks([])
            plt.gca().set_yticks([])

            plt.grid(True, alpha=0.3)

            plt.title(f"Bird Eye View - Step {self.current_step}")

            # Save figure
            if save_fig_dir is not None:
                os.makedirs(save_fig_dir, exist_ok=True)
                filename = f"bird-eye-view_step_{str(self.current_step).zfill(3)}.{image_format}"
                plt.savefig(
                    os.path.join(save_fig_dir, filename), dpi=150, bbox_inches="tight"
                )

            plt.show()

    def separate_overlapping_nodes(self, pos, min_distance=0.1, max_iterations=50):
        """Separate overlapping nodes by applying small adjustments to their positions."""
        import random

        import numpy as np

        nodes = list(pos.keys())
        adjusted_pos = pos.copy()

        for iteration in range(max_iterations):
            overlaps_found = False

            for i in range(len(nodes)):
                for j in range(i + 1, len(nodes)):
                    node1, node2 = nodes[i], nodes[j]
                    pos1 = np.array(adjusted_pos[node1])
                    pos2 = np.array(adjusted_pos[node2])

                    distance = np.linalg.norm(pos1 - pos2)

                    if distance < min_distance:
                        overlaps_found = True

                        # Calculate separation vector
                        if distance == 0:
                            # If nodes are exactly on top of each other, use random direction
                            angle = random.uniform(0, 2 * np.pi)
                            separation = (
                                np.array([np.cos(angle), np.sin(angle)]) * min_distance
                            )
                        else:
                            # Move nodes apart along the line connecting them
                            direction = (pos1 - pos2) / distance
                            separation = direction * (min_distance - distance) / 2

                        # Apply separation
                        adjusted_pos[node1] = pos1 + separation
                        adjusted_pos[node2] = pos2 - separation

            if not overlaps_found:
                break

        return adjusted_pos

    def _render_graph(
        self,
        figsize: tuple[int, int] = (12, 12),
        save_fig_dir: str = "./DEBUG/",
        image_format: str = "png",
        layout: str = "kamada_kawai",
    ) -> None:
        """Render the RDF graph as a network visualization."""
        # Get RDF graph and convert to networkx
        rdf_graph = self.get_rdf_graph()
        triples = rdf_to_list(rdf_graph)

        # Create networkx graph
        G = nx.DiGraph()

        # Add nodes and edges
        for subject, predicate, obj in triples:
            G.add_edge(subject, obj, label=predicate)

        # Create figure
        plt.figure(figsize=figsize)

        # Define node colors based on type
        node_colors = []
        color_mapping = {
            "agent": "#D8BFD8",  # Thistle (Soft Purple)
            "wall": "#D3D3D3",  # Light Gray
        }

        # Color rooms
        room_color = "#FFE4B5"  # Moccasin (Soft Gold)
        static_color = "#87CEFA"  # Light Blue
        moving_color = "#90EE90"  # Light Green

        for node in G.nodes():
            if node in self.room_names:
                node_colors.append(room_color)
            elif node in self.static_names:
                node_colors.append(static_color)
            elif node in self.moving_names:
                node_colors.append(moving_color)
            elif node == "agent":
                node_colors.append(color_mapping["agent"])
            elif node == "wall":
                node_colors.append(color_mapping["wall"])
            else:
                node_colors.append("#FFFFFF")  # White for unknown

        # Choose layout with better spacing
        if layout == "spring":
            pos = nx.spring_layout(G)
        elif layout == "circular":
            pos = nx.circular_layout(G)
        elif layout == "kamada_kawai":
            pos = nx.kamada_kawai_layout(G)
        else:
            raise ValueError(f"Unknown layout: {layout}.")

        # Separate overlapping nodes
        pos = self.separate_overlapping_nodes(
            pos, min_distance=0.15, max_iterations=100
        )

        # Draw nodes with larger size for better visibility
        nx.draw_networkx_nodes(
            G,
            pos,
            node_color=node_colors,
            node_size=3000,
            alpha=0.9,
            edgecolors="black",
            linewidths=2,
        )

        # Draw node labels with consistent styling
        nx.draw_networkx_labels(
            G, pos, font_size=12, font_weight="bold", font_color="black"
        )

        # Draw edges with better spacing
        nx.draw_networkx_edges(
            G,
            pos,
            edge_color="gray",
            arrows=True,
            arrowsize=25,
            arrowstyle="->",
            alpha=0.7,
            width=2,
            connectionstyle="arc3,rad=0.1",  # Slight curve to avoid overlap
        )

        # Draw edge labels with consistent black text and same font size
        edge_labels = nx.get_edge_attributes(G, "label")
        nx.draw_networkx_edge_labels(
            G,
            pos,
            edge_labels,
            font_size=12,
            font_color="black",
            font_weight="bold",
            alpha=1.0,
            bbox=dict(boxstyle="round,pad=0.2", facecolor="white", alpha=0.8),
        )

        plt.title(
            f"Graph View - Step {self.current_step}",
            fontsize=16,
            fontweight="bold",
        )
        plt.axis("off")
        plt.tight_layout()

        # Save figure
        if save_fig_dir is not None:
            os.makedirs(save_fig_dir, exist_ok=True)
            filename = (
                f"graph-view_step_{str(self.current_step).zfill(3)}.{image_format}"
            )
            plt.savefig(
                os.path.join(save_fig_dir, filename), dpi=150, bbox_inches="tight"
            )

        plt.show()

    def _draw_walls(self):
        """Draw outer walls and inner walls with thick black lines."""
        # Outer walls - very thick black
        wall_thickness = 8

        # Top wall
        plt.plot(
            [0, self.grid_length],
            [self.grid_length, self.grid_length],
            "k-",
            linewidth=wall_thickness,
        )
        # Bottom wall
        plt.plot([0, self.grid_length], [0, 0], "k-", linewidth=wall_thickness)
        # Left wall
        plt.plot([0, 0], [0, self.grid_length], "k-", linewidth=wall_thickness)
        # Right wall
        plt.plot(
            [self.grid_length, self.grid_length],
            [0, self.grid_length],
            "k-",
            linewidth=wall_thickness,
        )

        inner_wall_thickness = 4

        # Draw active inner walls based on current patterns
        for wall, pattern in self.wall_configs.items():
            pattern_length = len(pattern)
            pattern_index = self.current_step % pattern_length
            is_wall_active = pattern[pattern_index] == 1

            if is_wall_active:
                room1, room2, wall_type = wall
                pos1 = self.room_positions[room1]
                pos2 = self.room_positions[room2]

                if wall_type == "horizontal":
                    # Wall between vertically adjacent rooms
                    i1, j1 = pos1
                    i2, j2 = pos2
                    y = self.grid_length - max(i1, i2)  # Convert to plot coordinates
                    x_start = j1
                    x_end = j1 + 1
                    plt.plot(
                        [x_start, x_end], [y, y], "k-", linewidth=inner_wall_thickness
                    )

                else:  # vertical
                    # Wall between horizontally adjacent rooms
                    i1, j1 = pos1
                    i2, j2 = pos2
                    x = max(j1, j2)  # Convert to plot coordinates
                    y_start = self.grid_length - i1 - 1
                    y_end = self.grid_length - i1
                    plt.plot(
                        [x, x], [y_start, y_end], "k-", linewidth=inner_wall_thickness
                    )

    def get_rdf_graph(self):
        """Get current state as RDF graph."""
        g = Graph()

        # Add room connections
        for room_name, connections in self.room_connections.items():
            for direction, connected in connections.items():
                g.add((URIRef(room_name), URIRef(direction), URIRef(connected)))
        # Add object locations
        for obj_name, location in self.static_locations.items():
            g.add((URIRef(obj_name), URIRef("at_location"), URIRef(location)))

        for obj_name, location in self.moving_locations.items():
            g.add((URIRef(obj_name), URIRef("at_location"), URIRef(location)))

        # Add agent location
        g.add((URIRef("agent"), URIRef("at_location"), URIRef(self.agent_location)))

        return g



"""Room Creator for v3 - Simplified square grid with deterministic behavior."""

import random
from itertools import combinations

from .utils import read_lines
from .utils import write_json_prod as write_json


class RoomCreator:
    def __init__(
        self,
        filename: str = "dev",
        grid_length: int = 3,
        num_static_objects: int = 3,
        num_moving_objects: int = 3,
        num_inner_walls: int = 3,
        seed: int = 42,
    ) -> None:
        """Create a simplified square room environment.

        Args:
            filename: Config filename to save
            grid_length: Size of square grid (must be odd)
            num_static_objects: Number of static objects
            num_moving_objects: Number of moving objects
            num_inner_walls: Number of inner walls to select
            seed: Random seed for reproducibility
        """
        assert grid_length % 2 == 1, "grid_length must be odd"

        self.filename = filename
        self.grid_length = grid_length
        self.num_rooms = grid_length**2
        self.num_static_objects = num_static_objects
        self.num_moving_objects = num_moving_objects
        self.num_inner_walls = num_inner_walls
        self.seed = seed

        # Define the 10 possible wall patterns
        self.wall_patterns = [
            [1, 1, 1, 0, 1],
            [1, 1, 1, 1, 0, 1, 0, 1, 0, 0],
            [1, 0],
            [1, 1, 0, 1, 0, 1],
            [1, 1, 0, 0, 1],
            [1, 1, 0, 0],
            [1, 0, 0],
            [1, 1, 1, 0, 1, 0, 1, 0, 1],
            [0, 0, 0, 0, 0, 1, 1],
            [0, 1, 1, 1, 1, 0, 0, 1],
        ]

        random.seed(self.seed)

    def run(self):
        """Create and save the room configuration."""
        self._create_room_grid()
        self._create_room_connections()
        self._create_object_configs()
        self._create_wall_configs()
        self._create_question_list()
        self._save_config()

    def _create_room_grid(self):
        """Create square grid of rooms."""
        # Load names
        room_names = read_lines("data/room_names.txt")[: self.num_rooms]
        static_names = read_lines("data/static_objects.txt")[: self.num_static_objects]
        moving_names = read_lines("data/moving_objects.txt")[: self.num_moving_objects]

        self.room_names = room_names
        self.static_names = static_names
        self.moving_names = moving_names

        # Create grid positions
        self.room_positions = {}
        for i in range(self.grid_length):
            for j in range(self.grid_length):
                room_idx = i * self.grid_length + j
                self.room_positions[room_names[room_idx]] = (i, j)

    def _create_room_connections(self):
        """Create room connections with outer walls."""
        self.room_connections = {}

        for room_name, (i, j) in self.room_positions.items():
            connections = {
                "north": "wall",
                "east": "wall",
                "south": "wall",
                "west": "wall",
            }

            # North
            if i > 0:
                north_idx = (i - 1) * self.grid_length + j
                connections["north"] = self.room_names[north_idx]

            # East
            if j < self.grid_length - 1:
                east_idx = i * self.grid_length + (j + 1)
                connections["east"] = self.room_names[east_idx]

            # South
            if i < self.grid_length - 1:
                south_idx = (i + 1) * self.grid_length + j
                connections["south"] = self.room_names[south_idx]

            # West
            if j > 0:
                west_idx = i * self.grid_length + (j - 1)
                connections["west"] = self.room_names[west_idx]

            self.room_connections[room_name] = connections

    def _create_object_configs(self):
        """Create object initial locations and movement preferences."""
        # First, determine the center room where agent will start
        center_idx = self.grid_length // 2
        center_room = self.room_names[center_idx * self.grid_length + center_idx]
        self.agent_location = center_room

        # Available rooms exclude the center room (where agent starts)
        available_rooms = [room for room in self.room_names if room != center_room]

        # Static objects - ensure each is in a different room
        self.static_locations = {}
        for obj_name in self.static_names:
            room = random.choice(available_rooms)
            self.static_locations[obj_name] = room
            available_rooms.remove(room)

        # Moving objects - ensure each is in a different room from static objects and each other
        self.moving_locations = {}
        self.movement_preferences = {}
        directions = ["north", "east", "south", "west"]

        for obj_name in self.moving_names:
            room = random.choice(available_rooms)
            self.moving_locations[obj_name] = room
            available_rooms.remove(room)

            preferences = directions.copy()
            random.shuffle(preferences)
            self.movement_preferences[obj_name] = preferences

    def _set_agent_location(self):
        """Set agent starting location to center room."""
        # Calculate center position
        center_idx = self.grid_length // 2
        center_room = self.room_names[center_idx * self.grid_length + center_idx]
        self.agent_location = center_room

    def _create_wall_configs(self):
        """Create periodic inner wall configurations."""
        # Get all possible inner wall positions
        possible_walls = []

        # Horizontal walls (between vertically adjacent rooms)
        for i in range(self.grid_length - 1):
            for j in range(self.grid_length):
                room1 = i * self.grid_length + j
                room2 = (i + 1) * self.grid_length + j
                possible_walls.append(
                    (self.room_names[room1], self.room_names[room2], "horizontal")
                )

        # Vertical walls (between horizontally adjacent rooms)
        for i in range(self.grid_length):
            for j in range(self.grid_length - 1):
                room1 = i * self.grid_length + j
                room2 = i * self.grid_length + (j + 1)
                possible_walls.append(
                    (self.room_names[room1], self.room_names[room2], "vertical")
                )

        # Ensure we don't select more walls than possible
        max_walls = len(possible_walls)
        if self.num_inner_walls > max_walls:
            self.num_inner_walls = max_walls
            print(f"Warning: Reduced num_inner_walls to {max_walls} (maximum possible)")

        # Select walls that maintain connectivity
        self.selected_walls = []
        attempts = 0
        max_attempts = 1000

        while (
            len(self.selected_walls) < self.num_inner_walls and attempts < max_attempts
        ):
            # Randomly select remaining walls
            remaining_walls = [
                w for w in possible_walls if w not in self.selected_walls
            ]
            candidate_wall = random.choice(remaining_walls)

            # Test if adding this wall maintains connectivity
            test_walls = self.selected_walls + [candidate_wall]
            if self._test_connectivity_with_all_walls_on(test_walls):
                self.selected_walls.append(candidate_wall)

            attempts += 1

        if len(self.selected_walls) < self.num_inner_walls:
            print(
                f"Warning: Could only select {len(self.selected_walls)} walls that maintain connectivity"
            )

        # Assign patterns to selected walls - convert tuples to strings for JSON serialization
        self.wall_configs = {}
        for wall in self.selected_walls:
            pattern = random.choice(self.wall_patterns)
            # Convert tuple to string key for JSON compatibility
            wall_key = f"{wall[0]}|{wall[1]}|{wall[2]}"
            self.wall_configs[wall_key] = pattern

    def _test_connectivity_with_all_walls_on(self, walls):
        """Test if rooms remain connected when all specified walls are active."""
        # Create temporary room connections with all walls on
        temp_connections = {}
        for room_name in self.room_names:
            temp_connections[room_name] = self.room_connections[room_name].copy()

        # Apply all walls
        for room1, room2, wall_type in walls:
            if wall_type == "horizontal":
                temp_connections[room1]["south"] = "wall"
                temp_connections[room2]["north"] = "wall"
            else:  # vertical
                temp_connections[room1]["east"] = "wall"
                temp_connections[room2]["west"] = "wall"

        # Check connectivity using DFS
        visited = set()
        start_room = self.room_names[0]
        self._dfs(start_room, temp_connections, visited)

        return len(visited) == len(self.room_names)

    def _dfs(self, room, connections, visited):
        """Depth-first search for connectivity check."""
        visited.add(room)
        for direction in ["north", "east", "south", "west"]:
            neighbor = connections[room][direction]
            if neighbor != "wall" and neighbor not in visited:
                self._dfs(neighbor, connections, visited)

    def _create_question_list(self):
        """Create list of exactly 100 random questions."""
        all_objects = self.static_names + self.moving_names
        self.question_objects = [random.choice(all_objects) for _ in range(100)]

    def _save_config(self):
        """Save configuration to JSON file."""
        config = {
            "grid_length": self.grid_length,
            "room_names": self.room_names,
            "room_positions": self.room_positions,
            "room_connections": self.room_connections,
            "static_names": self.static_names,
            "moving_names": self.moving_names,
            "static_locations": self.static_locations,
            "moving_locations": self.moving_locations,
            "movement_preferences": self.movement_preferences,
            "agent_location": self.agent_location,
            "selected_walls": self.selected_walls,
            "wall_configs": self.wall_configs,
            "question_objects": self.question_objects,
            "seed": self.seed,
        }

        write_json(config, f"data/room-config-{self.filename}-v3.json")
        print(f"Saved configuration to room-config-{self.filename}-v3.json")
        print(f"Selected {len(self.selected_walls)} inner walls with periodic patterns")



import os
import random
from copy import deepcopy
from pprint import pprint

import gymnasium as gym
import matplotlib.pyplot as plt
from IPython.display import clear_output
# -----------------------------------------------------------------
from rdflib import Graph, URIRef

from ..utils import is_running_notebook
from ..utils import read_json_prod as read_json
from ..utils import sample_max_value_key, seed_everything

# -----------------------------------------------------------------


EPSILON = 1e-3


# NEW: helper function for converting an rdflib.Graph to a Python dict
def rdf_to_dict(g: Graph) -> dict:
    """Convert an rdflib Graph to a Python dictionary representation."""
    graph_dict = {}
    for s, p, o in g:
        s_key = str(s)  # Convert subject to string
        p_key = str(p)  # Convert predicate to string
        o_value = str(o)  # Convert object to string

        # Initialize subject entry if not present
        if s_key not in graph_dict:
            graph_dict[s_key] = {}

        # If predicate exists, store as a list (for multiple values)
        if p_key in graph_dict[s_key]:
            if isinstance(graph_dict[s_key][p_key], list):
                graph_dict[s_key][p_key].append(o_value)
            else:
                graph_dict[s_key][p_key] = [graph_dict[s_key][p_key], o_value]
        else:
            graph_dict[s_key][p_key] = o_value

    return graph_dict


class Object:
    """
    Base class representing any object in the room environment.

    This class serves as a parent for all object types (static, independent, dependent, agent).
    It handles basic properties like name, location, and movement logic common to all objects.

    Attributes:
        name (str): The unique identifier for the object.
        type (str): The type of object (static, independent, dependent, agent).
        init_probs (dict): Initial probabilities for starting locations.
        transition_probs (dict): Transition probabilities for movement.
        question_prob (float): Probability of being chosen for a question.
        deterministic (bool): Whether object behavior is deterministic.
        location (str): Current location of the object.
        history (list): History of locations the object has been in.
    """

    def __init__(
        self,
        name: str,
        type: str,
        init_probs: dict,
        transition_probs: dict,
        question_prob: float,
        deterministic: bool,
    ) -> None:
        """
        Initialize an Object.

        Args:
            name: Unique identifier for this object.
            type: Type of object (static, independent, dependent, agent).
            init_probs: Dictionary mapping room names to initial placement
                probabilities.
            transition_probs: Dictionary of movement transition probabilities.
            question_prob: Probability of being chosen for a question.
            deterministic: Whether behavior should be deterministic.

        Raises:
            ValueError: If initial probabilities don't sum to 1.
            AssertionError: If question_prob is not between 0 and 1.
        """
        self.name = name
        self.type = type
        self.init_probs = init_probs
        self.history = []

        if abs(sum(self.init_probs.values()) - 1) >= EPSILON:
            raise ValueError(
                f"The sum of the initial probabilities must be 1. "
                f"but it's {sum(self.init_probs.values())}"
            )
        self.transition_probs = transition_probs
        self.question_prob = question_prob
        self.deterministic = deterministic

        assert (
            0 <= self.question_prob <= 1
        ), f"question_prob must be between 0 and 1, but it's {self.question_prob}"

        # place an object in one of the rooms when it is created.
        if self.deterministic:
            self.location = sample_max_value_key(self.init_probs)
        else:
            self.location = random.choices(
                list(self.init_probs.keys()),
                weights=list(self.init_probs.values()),
                k=1,
            )[0]

    def __repr__(self) -> str:
        """
        Return a string representation of the object.

        Returns:
            String representation showing object type, name, and location.
        """
        return f"{self.type.title()}Object(name: {self.name}, location: {self.location}"

    def __eq__(self, other) -> bool:
        """
        Check equality between this object and another.

        Args:
            other: Another Object instance to compare with.

        Returns:
            True if objects have the same properties, False otherwise.
        """
        return (
            self.name == other.name
            and self.type == other.type
            and self.init_probs == other.init_probs
            and self.transition_probs == other.transition_probs
            and self.location == other.location
        )

    def move_with_action(self, action: str, rooms: dict, current_location: str) -> str:
        """
        Determine the next location based on a given action.

        Args:
            action: Direction to move ('north', 'east', 'south', 'west', 'stay').
            rooms: Dictionary of Room objects.
            current_location: Current location name.

        Returns:
            Name of the next location after taking the action.

        Raises:
            AssertionError: If action is not valid.
        """
        assert action in [
            "north",
            "east",
            "south",
            "west",
            "stay",
        ], f"{action} is not valid"
        if action == "north":
            next_location = rooms[current_location].north
        elif action == "east":
            next_location = rooms[current_location].east
        elif action == "south":
            next_location = rooms[current_location].south
        elif action == "west":
            next_location = rooms[current_location].west
        elif action == "stay":
            next_location = current_location

        return next_location if next_location != "wall" else current_location

    def _update_history(self) -> None:
        """
        Update the location history of this object.

        Appends the current location to the history list.
        """
        self.history.append(self.location)


class StaticObject(Object):
    """
    Represents a static object that stays in one place.

    Static objects do not move during the simulation.

    Attributes:
        Same as Object class, but transition_probs is always None.
    """

    def __init__(
        self,
        name: str,
        init_probs: dict,
        transition_probs: dict,
        question_prob: float,
        deterministic: bool,
    ) -> None:
        """
        Initialize a StaticObject.

        Args:
            name: Unique identifier for this object.
            init_probs: Dictionary mapping room names to initial placement probabilities.
            transition_probs: Must be None for static objects.
            question_prob: Probability of being chosen for a question.
            deterministic: Whether behavior should be deterministic.

        Raises:
            AssertionError: If transition_probs is not None.
        """
        super().__init__(
            name,
            "static",
            init_probs,
            transition_probs,
            question_prob,
            deterministic=deterministic,
        )
        # for a static object, we don't use transition_probs
        assert self.transition_probs is None, "Static objects do not move."

    def __repr__(self) -> str:
        """
        Return a string representation of the static object.

        Returns:
            String representation of the static object.
        """
        return super().__repr__() + ")"


class IndepdentObject(Object):
    """
    Represents an independent object that can move on its own.

    Independent objects move according to their transition probabilities
    and can have dependent objects attached to them.

    Attributes:
        rooms (dict): Dictionary of Room objects.
        attached (list): List of dependent objects attached to this object.
    """

    def __init__(
        self,
        name: str,
        init_probs: dict,
        transition_probs: dict,
        rooms: dict,
        question_prob: float,
        deterministic: bool,
    ) -> None:
        """
        Initialize an IndepdentObject.

        Args:
            name: Unique identifier for this object.
            init_probs: Dictionary mapping room names to initial placement probabilities.
            transition_probs: Dictionary of movement transition probabilities.
            rooms: Dictionary of Room objects.
            question_prob: Probability of being chosen for a question.
            deterministic: Whether behavior should be deterministic.

        Raises:
            ValueError: If transition probabilities for any state don't sum to 1.
        """
        super().__init__(
            name,
            "independent",
            init_probs,
            transition_probs,
            question_prob,
            deterministic,
        )
        for key, val in self.transition_probs.items():
            if abs(sum(val.values()) - 1) >= EPSILON:
                raise ValueError(
                    "The sum of the transition probabilities for an independent object "
                    f"must be 1. but it's {sum(val.values())}"
                )
        self.attached = []
        self.rooms = rooms

    def move(self) -> None:
        """
        Move the independent object according to its transition probabilities.

        Also moves any attached dependent objects to the same new location.
        """
        if self.deterministic:
            # pick the best action by max prob, ignoring "stay"
            action = sample_max_value_key(
                self.transition_probs[self.location], keys_to_exclude=["stay"]
            )
        else:
            action = random.choices(
                list(self.transition_probs[self.location].keys()),
                weights=list(self.transition_probs[self.location].values()),
                k=1,
            )[0]

        self.location = self.move_with_action(action, self.rooms, self.location)
        for do in self.attached:
            do.location = self.location
        self.detach()

    def detach(self) -> None:
        """
        Detach all dependent objects from this independent object.

        Clears the attachment relationships in both directions.
        """
        for do in self.attached:
            do.attached = None
        self.attached = []

    def __repr__(self) -> str:
        """
        Return a string representation of the independent object.

        Returns:
            String representation including attached objects.
        """
        return super().__repr__() + f", attached: {[do.name for do in self.attached]})"

    def __eq__(self, other) -> bool:
        """
        Check equality between this object and another.

        Args:
            other: Another IndepdentObject instance to compare with.

        Returns:
            True if objects have the same properties, False otherwise.
        """
        return (
            super().__eq__(other)
            and self.attached == other.attached
            and self.rooms == other.rooms
        )


class DependentObject(Object):
    """
    Represents a dependent object that can attach to independent objects.

    Dependent objects move when attached to an independent object that moves.
    They can attach to independent objects in the same room based on probabilities.

    Attributes:
        independent_objects (list): List of independent objects in the environment.
        attached (IndepdentObject or None): The independent object this object is attached to.
    """

    def __init__(
        self,
        name: str,
        init_probs: dict,
        transition_probs: dict,
        independent_objects: list,
        question_prob: float,
        deterministic: bool,
    ) -> None:
        """
        Initialize a DependentObject.

        Args:
            name: Unique identifier for this object.
            init_probs: Dictionary mapping room names to initial placement probabilities.
            transition_probs: Dictionary mapping independent object names to attachment probabilities.
            independent_objects: List of independent objects in the environment.
            question_prob: Probability of being chosen for a question.
            deterministic: Whether behavior should be deterministic.

        Raises:
            ValueError: If any transition probability is greater than 1.
        """
        super().__init__(
            name,
            "dependent",
            init_probs,
            transition_probs,
            question_prob,
            deterministic,
        )
        for key, val in self.transition_probs.items():
            if val >= 1 + EPSILON:
                raise ValueError(
                    "The transition probability for a dependent object must "
                    f"be <= 1. but it's {val}"
                )
        self.independent_objects = independent_objects
        self.attach()

    def attach(self) -> None:
        """
        Try to attach to an independent object in the same room.

        The attachment is probabilistic based on transition_probs unless deterministic is True.
        """
        self.attached = None
        possible_attachments = []
        for io in self.independent_objects:
            if io.location == self.location:
                for io_name, prob in self.transition_probs.items():
                    if io.name == io_name:
                        if self.deterministic:
                            possible_attachments.append(io)
                        else:
                            if random.random() < prob:
                                possible_attachments.append(io)

        if len(possible_attachments) > 0:
            if self.deterministic:
                io = possible_attachments[0]
            else:
                io = random.choice(possible_attachments)
            self.attached = io
            if self.name not in [do.name for do in io.attached]:
                io.attached.append(self)

    def __repr__(self) -> str:
        """
        Return a string representation of the dependent object.

        Returns:
            String representation including attachment information.
        """
        if self.attached is None:
            return super().__repr__() + ", attached: None)"
        else:
            return super().__repr__() + f", attached: {self.attached.name})"

    def __eq__(self, other) -> bool:
        """
        Check equality between this object and another.

        Args:
            other: Another DependentObject instance to compare with.

        Returns:
            True if objects have the same properties, False otherwise.
        """
        return (
            super().__eq__(other)
            and self.attached == other.attached
            and self.independent_objects == other.independent_objects
        )


class Agent(Object):
    """
    Represents an agent that can be controlled through actions.

    The agent moves according to external actions rather than transition probabilities.

    Attributes:
        rooms (dict): Dictionary of Room objects.
    """

    def __init__(
        self,
        name: str,
        init_probs: dict,
        transition_probs: dict,
        rooms: dict,
        question_prob: float,
    ) -> None:
        """
        Initialize an Agent.

        Args:
            name: Unique identifier for this agent.
            init_probs: Dictionary mapping room names to initial placement probabilities.
            transition_probs: Must be None for agents.
            rooms: Dictionary of Room objects.
            question_prob: Must be close to 0 as agents are not questionable.

        Raises:
            AssertionError: If transition_probs is not None or question_prob is not close to 0.
        """
        assert abs(question_prob) <= EPSILON, "Agents are not questionable."

        super().__init__(
            name,
            "agent",
            init_probs,
            transition_probs,
            question_prob,
            deterministic=False,
        )
        assert self.transition_probs is None, "Agent objects do not move by itself."
        self.rooms = rooms

    def move(self, action: str) -> None:
        """
        Move the agent according to the specified action.

        Args:
            action: Direction to move ('north', 'east', 'south', 'west', 'stay').
        """
        self.location = self.move_with_action(action, self.rooms, self.location)

    def __repr__(self) -> str:
        """
        Return a string representation of the agent.

        Returns:
            String representation of the agent.
        """
        return "Agent(name: agent, location: " + self.location + ")"

    def __eq__(self, other) -> bool:
        """
        Check equality between this agent and another.

        Args:
            other: Another Agent instance to compare with.

        Returns:
            True if agents have the same properties, False otherwise.
        """
        return super().__eq__(other) and self.rooms == other.rooms


class Room:
    """
    Represents a room in the environment.

    Rooms are connected to each other in the four cardinal directions.

    Attributes:
        name (str): Unique name of the room.
        north (str): Name of the room to the north, or 'wall'.
        east (str): Name of the room to the east, or 'wall'.
        south (str): Name of the room to the south, or 'wall'.
        west (str): Name of the room to the west, or 'wall'.
    """

    def __init__(self, name: str, north: str, east: str, south: str, west: str) -> None:
        """
        Initialize a Room.

        Args:
            name: Unique name for this room.
            north: Name of the room to the north, or 'wall'.
            east: Name of the room to the east, or 'wall'.
            south: Name of the room to the south, or 'wall'.
            west: Name of the room to the west, or 'wall'.

        Raises:
            AssertionError: If the room layout is invalid (e.g., duplicate connections).
        """
        self.name = name
        self.north = north
        self.east = east
        self.south = south
        self.west = west

        rooms_walls = [self.north, self.east, self.south, self.west]
        rooms_walls = [rw for rw in rooms_walls if rw != "wall"]
        assert len(set(rooms_walls)) == len(rooms_walls), "room layout wrong."

    def __repr__(self) -> str:
        """
        Return a string representation of the room.

        Returns:
            String representation showing room name and connections.
        """
        return (
            f"Room(name: {self.name}, north: {self.north}, east: {self.east}, "
            f"south: {self.south}, west: {self.west})"
        )

    def __eq__(self, other) -> bool:
        """
        Check equality between this room and another.

        Args:
            other: Another Room instance to compare with.

        Returns:
            True if rooms have the same properties, False otherwise.
        """
        return (
            self.name == other.name
            and self.north == other.north
            and self.east == other.east
            and self.south == other.south
            and self.west == other.west
        )


class RoomEnv2(gym.Env):
    """
    The Room environment version 2.

    A gymnasium environment for simulating objects and an agent in interconnected rooms.
    Objects can move independently or be attached to other objects, and the agent can be
    controlled through actions. The environment supports question-answering about object locations.

    Attributes:
        rooms (dict): Dictionary of Room objects.
        objects (dict): Dictionary containing lists of objects by type.
        relations (list): List of relation types in the environment.
        entities (dict): Dictionary of entity names by type.
        hidden_global_state (list): Current state as triples.
        observations_room (list): Current observations of the room.
        questions (list): Current questions.
        answers (list): Current answers to questions.
    """

    def __init__(
        self,
        question_prob: int = 1.0,
        seed: int = 42,
        terminates_at: int = 99,
        randomize_observations: str = "all",
        room_size: str = "xl-different-prob",
        rewards: dict = {"correct": 1, "wrong": 0, "partial": 0},
        make_everything_static: bool = False,
        num_total_questions: int = 1000,
        question_interval: int = 1,
        include_walls_in_observations: bool = True,
        deterministic_objects: bool = False,
    ) -> None:
        """
        Initialize the Room environment.

        Args:
            question_prob: Probability of generating a question when appropriate.
            seed: Random seed for reproducibility.
            terminates_at: Number of steps before the episode terminates.
            randomize_observations: How to randomize observations ('all', 'objects', 'objects_middle', 'none').
            room_size: Size configuration of the room layout.
            rewards: Dictionary specifying rewards for correct, wrong, and partial answers.
            make_everything_static: If True, objects don't move.
            num_total_questions: Total number of questions to generate throughout all episodes.
            question_interval: Interval between generating questions.
            include_walls_in_observations: Whether to include walls in observations.
            deterministic_objects: Whether objects should behave deterministically.

        Raises:
            AssertionError: If the total questions or terminates_at don't align with question_interval.
        """
        super().__init__()
        self.is_notebook = is_running_notebook()
        if isinstance(room_size, str):
            config_all = read_json(f"./data/room-config-{room_size}-v2.json")
        else:
            for key in [
                "object_init_config",
                "object_transition_config",
                "room_config",
            ]:
                assert key in room_size, f"{key} is not in the room_size dict."
            config_all = room_size

        self.room_config = config_all["room_config"]
        if "mapping" in config_all:
            self.mapping = config_all["mapping"]
        self.object_transition_config = config_all["object_transition_config"]
        self.object_init_config = config_all["object_init_config"]
        self.object_question_probs = config_all["object_question_probs"]

        if "grid" in config_all:
            self.grid = config_all["grid"]
        if "room_indexes" in config_all:
            self.room_indexes = config_all["room_indexes"]
        if "names" in config_all:
            self.names = config_all["names"]

        self.seed = seed
        seed_everything(self.seed)
        self.question_prob = question_prob
        self.terminates_at = terminates_at
        self.randomize_observations = randomize_observations
        self.num_total_questions = num_total_questions
        self.question_interval = question_interval
        self.include_walls_in_observations = include_walls_in_observations
        self.deterministic_objects = deterministic_objects

        if self.deterministic_objects:
            assert (
                self.randomize_observations == "none"
            ), "Deterministic objects should not have randomized observations."

        assert self.num_total_questions % (self.terminates_at + 1) == 0, (
            f"The total number of questions must be a multiple of "
            f"{self.terminates_at + 1}, but it's {self.num_total_questions}"
        )
        assert (self.terminates_at + 1) % self.question_interval == 0, (
            f"The total number of steps must be a multiple of "
            f"{self.question_interval}, but it's {self.terminates_at + 1}"
        )

        self.num_questions_step = (
            self.num_total_questions
            // (self.terminates_at + 1)
            * self.question_interval
        )
        self.total_maximum_episode_rewards = num_total_questions

        self._create_rooms()
        self._compute_room_map()
        self._create_objects()
        self._create_relations_and_objects_for_nn()

        # placeholders for gym spaces
        self.observation_space = gym.spaces.Discrete(1)
        self.action_space = gym.spaces.Discrete(1)

        self.CORRECT, self.WRONG, self.PARTIAL = (
            rewards["correct"],
            rewards["wrong"],
            rewards["partial"],
        )
        self.make_everything_static = make_everything_static

        self._init_lists()

    def _init_lists(self) -> None:
        self.hidden_global_states_all = []
        self.observations_all = []
        self.answers_all = []
        self.info_all = []

    def _create_rooms(self) -> None:
        """
        Create room objects based on the room configuration.

        Populates self.rooms with Room objects.
        """
        self.rooms = {}
        for name, config_ in self.room_config.items():
            self.rooms[name] = Room(name, **config_)

    def _create_objects(self) -> None:
        """
        Create all objects (static, independent, dependent, agent) based on configuration.

        Populates self.objects with various object instances.

        Raises:
            AssertionError: If the sum of question probabilities exceeds 1.
        """
        self.objects = {"static": [], "independent": [], "dependent": [], "agent": []}

        for name, init_probs in self.object_init_config["static"].items():
            self.objects["static"].append(
                StaticObject(
                    name,
                    init_probs,
                    self.object_transition_config["static"][name],
                    self.object_question_probs["static"][name],
                    self.deterministic_objects,
                )
            )

        for name, init_probs in self.object_init_config["independent"].items():
            self.objects["independent"].append(
                IndepdentObject(
                    name,
                    init_probs,
                    self.object_transition_config["independent"][name],
                    self.rooms,
                    self.object_question_probs["independent"][name],
                    self.deterministic_objects,
                )
            )

        for name, init_probs in self.object_init_config["dependent"].items():
            self.objects["dependent"].append(
                DependentObject(
                    name,
                    init_probs,
                    self.object_transition_config["dependent"][name],
                    self.objects["independent"],
                    self.object_question_probs["dependent"][name],
                    self.deterministic_objects,
                )
            )

        for name, init_probs in self.object_init_config["agent"].items():
            self.objects["agent"].append(
                Agent(
                    name,
                    init_probs,
                    self.object_transition_config["agent"][name],
                    self.rooms,
                    self.object_question_probs["agent"][name],
                )
            )

        # sanity-check sum of question probs
        question_probs = []
        for obj_type, objects in self.objects.items():
            for obj in objects:
                question_probs.append(obj.question_prob)
        assert abs(sum(question_probs) - 1) <= EPSILON, (
            "The sum of the question probabilities must be <= 1. but it's "
            f"{sum(question_probs)}"
        )

    def _create_relations_and_objects_for_nn(self) -> None:
        """
        Create dictionaries of relations and entities for use in neural networks.

        Populates self.relations and self.entities.
        """
        self.relations = ["north", "east", "south", "west", "at_location"]

        self.entities = {
            "static": [],
            "independent": [],
            "dependent": [],
            "agent": [],
            "room": [],
            "others": ["wall"],
        }
        for obj_type, objects in self.objects.items():
            for obj in objects:
                self.entities[obj_type].append(obj.name)

        for room_name in self.rooms:
            self.entities["room"].append(room_name)

    def _compute_room_map(self) -> None:
        """
        Compute the room layout map as a list of triples.

        Populates self.room_layout.
        """
        self.room_layout = []
        for name, room in self.rooms.items():
            self.room_layout.append([name, "north", room.north])
            self.room_layout.append([name, "east", room.east])
            self.room_layout.append([name, "south", room.south])
            self.room_layout.append([name, "west", room.west])

    def return_room_layout(self, exclude_walls: bool = False) -> list[list[str]]:
        """
        Return the room layout as a list of [room, direction, connected_room] triples.

        Args:
            exclude_walls: If True, exclude connections to walls.

        Returns:
            List of triples representing room connections.
        """
        room_layout = []
        for triple in self.room_layout:
            if exclude_walls:
                if triple[2] != "wall":
                    room_layout.append(triple)
            else:
                room_layout.append(triple)
        return room_layout

    def _compute_hidden_global_state(self) -> None:
        """
        Compute the hidden global state as [head, relation, tail] triples.

        Updates self.hidden_global_state and appends to self.hidden_global_states_all.
        """
        self.hidden_global_state = []

        # room adjacency
        for name, room in self.rooms.items():
            self.hidden_global_state.append([name, "north", room.north])
            self.hidden_global_state.append([name, "east", room.east])
            self.hidden_global_state.append([name, "south", room.south])
            self.hidden_global_state.append([name, "west", room.west])

        # object locations
        for obj_type in ["agent", "static", "independent", "dependent"]:
            for obj in self.objects[obj_type]:
                self.hidden_global_state.append([obj.name, "at_location", obj.location])

        self.hidden_global_states_all.append(deepcopy(self.hidden_global_state))

    def _find_object_by_string(self, obj_str: str) -> Object:
        """
        Find an object by its name.

        Args:
            obj_str: Name of the object to find.

        Returns:
            Object instance if found, None otherwise.
        """
        for obj_type, objects in self.objects.items():
            for obj in objects:
                if obj.name == obj_str:
                    return obj
        return None

    def get_observations_and_question(self, generate_questions: bool = True) -> dict:
        """
        Get observations about the current room and generate questions if required.

        Args:
            generate_questions: Whether to generate questions.

        Returns:
            Dictionary containing observations and questions.
        """
        agent_location = self.objects["agent"][0].location
        self._compute_hidden_global_state()
        self.observations_room = []

        for triple in self.hidden_global_state:
            subj, rel, obj_ = triple
            # show adjacency if subj == agent_location
            if rel in ["north", "east", "south", "west"] and subj == agent_location:
                if self.include_walls_in_observations or (obj_ != "wall"):
                    self.observations_room.append(triple)
            elif rel == "at_location":
                if obj_ == agent_location:  # show objects in same room as agent
                    self.observations_room.append(triple)

        # possibly shuffle the object portion
        if self.randomize_observations == "all":
            random.shuffle(self.observations_room)
        elif self.randomize_observations == "objects":
            first = [
                obs
                for obs in self.observations_room
                if obs[1] in ["north", "east", "south", "west"] or obs[0] == "agent"
            ]
            second = [obs for obs in self.observations_room if obs not in first]
            random.shuffle(second)
            self.observations_room = first + second

        elif self.randomize_observations == "objects_middle":
            first = [
                obs
                for obs in self.observations_room
                if obs[1] in ["north", "east", "south", "west"]
            ]
            third = [obs for obs in self.observations_room if obs[0] == "agent"]
            second = [obs for obs in self.observations_room if obs not in first + third]
            random.shuffle(second)
            self.observations_room = first + second + third
        elif self.randomize_observations == "none":
            pass

        # set up possible questions
        self.questions = []
        self.answers = []

        if generate_questions:
            question_candidates = [
                (obj.name, obj.question_prob)
                for obj_type, objs in self.objects.items()
                for obj in objs
            ]
            names, probs = zip(*question_candidates)
            chosen_tuples = random.choices(
                question_candidates, weights=probs, k=self.num_questions_step
            )
            names = [ct[0] for ct in chosen_tuples]

            objs_chosen = [self._find_object_by_string(name) for name in names]
            for i in range(self.num_questions_step):
                if random.random() > self.question_prob:
                    self.questions.append(None)
                    self.answers.append(None)
                else:
                    obj_chosen = objs_chosen[i]
                    self.questions.append([obj_chosen.name, "at_location", "?"])
                    answer = {"current": obj_chosen.location, "previous": None}

                    for prev_loc in reversed(obj_chosen.history):
                        if prev_loc != obj_chosen.location:
                            answer["previous"] = prev_loc
                            break
                    self.answers.append(answer)

        observations = deepcopy(
            {"room": self.observations_room, "questions": self.questions}
        )
        self.observations_all.append(observations)
        self.answers_all.append(self.answers)

        return observations

    def reset(self) -> tuple[dict, dict]:
        """
        Reset the environment to an initial state.

        Returns:
            Tuple of (observations, info_dict).
        """
        self._init_lists()
        info = {}
        self._create_rooms()
        self._create_objects()
        self.current_time = 0
        self.info_all.append(info)

        # initialize object histories
        for obj_type, objs in self.objects.items():
            for obj in objs:
                obj._update_history()

        # produce observation
        if (self.current_time + 1) % self.question_interval == 0:
            return self.get_observations_and_question(generate_questions=True), info
        else:
            return self.get_observations_and_question(generate_questions=False), info

    def step(
        self, actions: tuple[list[str], str]
    ) -> tuple[dict, int, bool, bool, dict]:
        """
        Take a step in the environment based on provided actions.

        Args:
            actions: Tuple of (question_answers, exploration_action).

        Returns:
            Tuple of (observations, rewards, done, truncated, info_dict).
        """
        actions_qa, action_explore = actions
        assert isinstance(actions_qa, list), "actions_qa must be a list."
        assert isinstance(action_explore, str), "action_explore must be a string."

        # scoring answers
        if len(self.answers) == 0:
            assert actions_qa == [], "You shouldn't answer any questions now"
            rewards = []
        else:
            assert len(actions_qa) == len(self.answers), "You must answer all questions"
            rewards = []
            for user_answer, true_answer in zip(actions_qa, self.answers):
                if user_answer == true_answer["current"]:
                    rewards.append(self.CORRECT)
                elif user_answer == true_answer["previous"]:
                    rewards.append(self.PARTIAL)
                else:
                    rewards.append(self.WRONG)

        # move environment forward
        if not self.make_everything_static:
            for obj in self.objects["independent"]:
                obj.move()
            for obj in self.objects["dependent"]:
                obj.attach()

        # agent action
        self.objects["agent"][0].move(action_explore)

        done = self.current_time == self.terminates_at
        truncated = False
        info = deepcopy({"answers": self.answers, "timestamp": self.current_time})
        self.info_all.append(info)

        self.current_time += 1

        for obj_type, objs in self.objects.items():
            for obj in objs:
                obj._update_history()

        # produce next observation
        if (self.current_time + 1) % self.question_interval == 0:
            obs = self.get_observations_and_question(generate_questions=True)
        else:
            obs = self.get_observations_and_question(generate_questions=False)

        return obs, rewards, done, truncated, info

    def _find_objects_in_room(self, room_name: str) -> dict[str, list]:
        """
        Find all objects currently in a specified room.

        Args:
            room_name: Name of the room to search.

        Returns:
            Dictionary mapping object types to lists of object names.
        """
        objects_in_room = {k: [] for k in self.objects.keys()}
        for obj_type, objs in self.objects.items():
            for obj in objs:
                if obj.location == room_name:
                    objects_in_room[obj_type].append(obj.name)
        return objects_in_room

    def render(
        self,
        render_mode: str = "console",
        figsize: tuple[int, int] = (15, 15),
        cell_text_size: int = 10,
        save_fig_dir: str = None,
        image_format: str = "png",
    ) -> None:
        """
        Render the current state of the environment.

        Args:
            render_mode: How to render the environment ('console' or 'image').
            figsize: Size of the figure when rendering as an image.
            cell_text_size: Text size in cells when rendering as an image.
            save_fig_dir: Directory to save figures if not None.
            image_format: Format to save images in.
        """
        if render_mode == "console":
            pprint(self.hidden_global_state)
        elif render_mode == "image":
            if self.is_notebook:
                clear_output(True)
            plt.figure(figsize=figsize)
            num_rows = len(self.grid)
            num_cols = len(self.grid[0])

            plt.subplot(111)
            plt.title(f"Hidden state at time={self.current_time}")

            for row in range(num_rows):
                for col in range(num_cols):
                    text = ""
                    cell_content = self.grid[row][col]
                    if cell_content != 0:
                        color = "white"
                        room_index = self.room_indexes.index([row, col])
                        room_name = self.names["room"][room_index]
                        text += f"room name={room_name}"

                        objects_in_room = self._find_objects_in_room(room_name)
                        for obj_type, objects in objects_in_room.items():
                            if objects:
                                text += f"\n{obj_type} objects: {objects}"
                    else:
                        color = "black"
                    plt.gca().add_patch(
                        plt.Rectangle((col, num_rows - 1 - row), 1, 1, facecolor=color)
                    )
                    plt.text(
                        col + 0.5,
                        num_rows - 1 - row + 0.5,
                        text,
                        ha="center",
                        va="center",
                        color="black",
                        fontsize=cell_text_size,
                    )
            plt.gca().set_aspect("equal")
            plt.gca().set_xticks(range(num_cols + 1))
            plt.gca().set_yticks(range(num_rows + 1))
            plt.gca().grid(which="both")

            if save_fig_dir is not None:
                os.makedirs(save_fig_dir, exist_ok=True)
                plt.savefig(
                    os.path.join(
                        save_fig_dir,
                        f"hidden_state-{str(self.current_time).zfill(3)}.{image_format}",
                    )
                )
            plt.show()

    def get_rdf_graph_from_state(self, state=None) -> Graph:
        """
        Build an rdflib.Graph from a list of [subject, predicate, object] triples.

        Args:
            state: List of triples to convert. If None, uses current
            hidden_global_state.

        Returns:
            rdflib.Graph representing the state.
        """
        if state is None:
            state = self.hidden_global_state

        g = Graph()

        for s, p, o in state:
            # Convert strings directly to URIRefs without namespace
            g.add((URIRef(s), URIRef(p), URIRef(o)))

        return g

    def get_rdf_graph_from_observations(self) -> Graph:
        """
        Build an rdflib.Graph from the current observations.

        Returns:
            rdflib.Graph representing the current observations.
        """
        g = Graph()

        # Add room observations
        for s, p, o in self.observations_room:
            g.add((URIRef(s), URIRef(p), URIRef(o)))

        return g

    def get_current_observations_rdf_dict(self) -> dict:
        """
        Return a Python dict representing the current observations as RDF.

        Returns:
            Dictionary representation of the current observations as RDF.
        """
        g = self.get_rdf_graph_from_observations()
        return rdf_to_dict(g)

    def get_current_hidden_state_rdf_dict(self) -> dict:
        """
        Return a Python dict representing the current hidden_global_state as RDF.

        Returns:
            Dictionary representation of the current RDF state.
        """
        g = self.get_rdf_graph_from_state(self.hidden_global_state)
        return rdf_to_dict(g)

    def get_all_rdf_dicts(self) -> list[dict]:
        """
        Return a list of Python dicts, one for each time-step's hidden_global_state.

        Returns:
            List of dictionary representations of all historical RDF states.
        """
        all_rdf_dicts = []
        for state in self.hidden_global_states_all:
            g = self.get_rdf_graph_from_state(state)
            all_rdf_dicts.append(rdf_to_dict(g))
        return all_rdf_dicts



import unittest

from room_env.des import RoomDes


class DesTest(unittest.TestCase):
    def test_all(self) -> None:
        for des_size in ["xxs", "xs", "s", "m", "l"]:
            des = RoomDes(des_size=des_size)
            des.run()



"""Only run this in Juputer notebook."""

import random
from copy import deepcopy
from pprint import pprint

import matplotlib.pyplot as plt
import numpy as np

from .utils import find_connected_nodes
from .utils import read_json_prod as read_json
from .utils import write_json_prod as write_json

EPSILON = 1e-3


class RoomCreator:
    def __init__(
        self,
        grid_length: int = 1,
        num_rooms: int = 1,
        num_static_objects: int = 1,
        num_independent_objects: int = 1,
        num_dependent_objects: int = 1,
        room_prob: float = 0.7,
        minimum_transition_stay_prob: float = 0.5,
        give_fake_names: bool = False,
        static_object_in_every_room: bool = True,
        filename: str = "dev",
        same_question_probs: bool = True,
    ) -> None:
        """Create rooms with objects.

        Args:
            grid_length: grid legnth
            num_rooms: number of rooms in the grid
            num_static_objects: Number of static objects to create.
            num_independent_objects: Number of independent objects to create.
            num_dependent_objects: Number of dependent objects to create.
            room_prob: probability of a cell being a room.
            minimum_transition_stay_prob: probability of staying in the same room.
                This should be quite high. Otherwise, the objects are not tractable.
            give_fake_names: If True, give fake names to the rooms and objects.
            static_object_in_every_room: If True, every room has at least one static
                object.
            filename: filename to save the config.
            same_question_probs: If True, all objects have the same probability of
                being asked.

        """
        self.filename = filename
        print(
            f"The config will be saved as "
            f"./room_env/data/room-config-{self.filename}-v2.json\n"
        )
        self.grid_length = grid_length
        self.num_rooms = num_rooms
        self.num_static_objects = num_static_objects
        self.num_independent_objects = num_independent_objects
        self.num_dependent_objects = num_dependent_objects
        self.room_prob = room_prob
        self.give_fake_names = give_fake_names
        self.minimum_transition_stay_prob = minimum_transition_stay_prob
        self.static_object_in_every_room = static_object_in_every_room
        self.same_question_probs = same_question_probs

    def run(self):
        self._create_grid_world()
        self._create_room_config()
        self._create_object_init_config()
        self._create_object_transition_config()
        self._give_names()
        self._create_question_probs()
        write_json(
            {
                "room_config": self.room_config_str,
                "object_init_config": self.object_init_config_str,
                "object_transition_config": self.object_transition_config_str,
                "object_question_probs": self.question_probs,
                "grid": self.grid,
                "room_indexes": self.room_indexes,
                "names": self.names,
            },
            f"./data/room-config-{self.filename}-v2.json",
        )

        self._visualize_grids()

    def _create_grid_world(self) -> None:
        rows = self.grid_length
        cols = self.grid_length

        while True:
            self.grid = [
                [1 if random.random() < self.room_prob else 0 for _ in range(cols)]
                for _ in range(rows)
            ]
            if sum(sum(row) for row in self.grid) == 0:
                continue

            self.connected_components = find_connected_nodes(self.grid)

            self.previous_grid = deepcopy(self.grid)

            self.room_indexes = max(
                enumerate(self.connected_components), key=lambda x: len(x[1])
            )[1]
            self.room_indexes.sort(key=lambda x: x[1])
            self.room_indexes.sort(key=lambda x: x[0])
            self.grid = [[0 for _ in range(cols)] for _ in range(rows)]
            for i in range(rows):
                for j in range(cols):
                    if (i, j) in self.room_indexes:
                        self.grid[i][j] = 1

            # This is to assure that every room as at least one static object.
            if len(self.room_indexes) == self.num_rooms:
                if self.static_object_in_every_room:
                    if self.num_static_objects >= self.num_rooms:
                        break
                else:
                    break

    def _create_room_config(self) -> None:
        """Create a room configuration."""
        self.room_config = {
            i: {"north": "wall", "east": "wall", "south": "wall", "west": "wall"}
            for i in range(self.num_rooms)
        }
        for i, (row_idx, col_idx) in enumerate(self.room_indexes):
            north_idx = (row_idx - 1, col_idx)
            east_idx = (row_idx, col_idx + 1)
            south_idx = (row_idx + 1, col_idx)
            west_idx = (row_idx, col_idx - 1)

            if north_idx in self.room_indexes:
                self.room_config[i]["north"] = self.room_indexes.index(north_idx)
            if east_idx in self.room_indexes:
                self.room_config[i]["east"] = self.room_indexes.index(east_idx)
            if south_idx in self.room_indexes:
                self.room_config[i]["south"] = self.room_indexes.index(south_idx)
            if west_idx in self.room_indexes:
                self.room_config[i]["west"] = self.room_indexes.index(west_idx)

    def _create_object_init_config(self) -> None:
        """Create an object initialization configuration."""
        self.object_init_config = {}

        # This is made to be dertministic.
        for object_type, num_objects in zip(
            ["static", "independent", "dependent", "agent"],
            [
                self.num_static_objects,
                self.num_independent_objects,
                self.num_dependent_objects,
                1,
            ],
        ):
            self.object_init_config[object_type] = {
                object_num: {room_num: 0 for room_num in range(self.num_rooms)}
                for object_num in range(num_objects)
            }

        for object_num in range(self.num_static_objects):
            if self.static_object_in_every_room and object_num < self.num_rooms:
                room_num = object_num
                self.object_init_config["static"][object_num][room_num] = 1
            else:
                room_num = random.randint(0, self.num_rooms - 1)
                self.object_init_config["static"][object_num][room_num] = 1

        for object_type, num_objects in zip(
            ["independent", "dependent"],
            [
                self.num_independent_objects,
                self.num_dependent_objects,
            ],
        ):
            for object_num in range(num_objects):
                room_num = random.randint(0, self.num_rooms - 1)
                self.object_init_config[object_type][object_num][room_num] = 1

        self.object_init_config["agent"][0][0] = 1

        for object_type in ["static", "independent", "dependent", "agent"]:
            for _, room_num_dist in self.object_init_config[object_type].items():
                assert (
                    abs(sum(room_num_dist.values()) - 1) < EPSILON
                ), f"{sum(room_num_dist.values())}"

    def _create_object_transition_config(self) -> None:
        """Create an object transition configuration."""
        self.object_transition_config = {}
        self.object_transition_config["static"] = {
            i: None for i in range(self.num_static_objects)
        }
        self.object_transition_config["independent"] = {
            object_num: {
                room_num: {
                    "north": 0,
                    "east": 0,
                    "south": 0,
                    "west": 0,
                    "west": 0,
                    "stay": 0,
                }
                for room_num in range(self.num_rooms)
            }
            for object_num in range(self.num_independent_objects)
        }
        for object_num, room_num_nesws in self.object_transition_config[
            "independent"
        ].items():
            for room_num, nesw in room_num_nesws.items():
                nesw["stay"] = random.uniform(self.minimum_transition_stay_prob, 1)

                for direction, _ in nesw.items():
                    if (
                        direction != "stay"
                        and self.room_config[room_num][direction] != "wall"
                    ):
                        nesw[direction] = np.random.beta(0.5, 0.5, 1).item()

                denominator = sum(
                    [num for direction, num in nesw.items() if direction != "stay"]
                )

                denominator /= 1 - nesw["stay"]

                if denominator != 0:
                    for direction, _ in nesw.items():
                        if direction != "stay":
                            nesw[direction] = nesw[direction] / denominator
                else:
                    nesw["stay"] = 1

        self.object_transition_config["dependent"] = {
            dep_object_num: {
                ind_object_num: np.random.beta(0.5, 0.5, 1).item()
                for ind_object_num in range(self.num_independent_objects)
            }
            for dep_object_num in range(self.num_dependent_objects)
        }
        self.object_transition_config["agent"] = {"agent": None}

    def _give_names(self) -> None:
        """Give (fake) names to the rooms and objects."""

        self.names = {}
        if self.give_fake_names:
            fake_names = read_json("./data/names-v2.json")

            for object_type, num_objects in zip(
                ["room", "static_objects", "independent_objects", "dependent_objects"],
                [
                    self.num_rooms,
                    self.num_static_objects,
                    self.num_independent_objects,
                    self.num_dependent_objects,
                ],
            ):
                assert num_objects <= len(fake_names[object_type]), (
                    f"{num_objects} should be lower than or equal to "
                    f"{len(fake_names[object_type])}"
                )

                self.names[object_type] = random.sample(
                    fake_names[object_type], k=num_objects
                )
        else:
            for object_type, num_objects in zip(
                ["room", "static_objects", "independent_objects", "dependent_objects"],
                [
                    self.num_rooms,
                    self.num_static_objects,
                    self.num_independent_objects,
                    self.num_dependent_objects,
                ],
            ):
                if object_type == "room":
                    string = "room"
                elif object_type == "static_objects":
                    string = "sta"
                elif object_type == "independent_objects":
                    string = "ind"
                elif object_type == "dependent_objects":
                    string = "dep"
                self.names[object_type] = [
                    string + "_" + str(object_num).zfill(3)
                    for object_num in range(num_objects)
                ]

        print(
            f"# of rooms: {len(self.names['room'])}\n"
            f"# of static objects: {len(self.names['static_objects'])} \n"
            f"# of independent objects: {len(self.names['independent_objects'])} \n"
            f"# of dependent objects: {len(self.names['dependent_objects'])} \n"
        )
        pprint(self.names)

        self.room_config_str = {}
        for room_num in range(self.num_rooms):
            self.room_config_str[self.names["room"][room_num]] = self.room_config[
                room_num
            ]

            self.room_config_str[self.names["room"][room_num]] = {
                direction: (
                    self.names["room"][room_num_or_wall]
                    if room_num_or_wall != "wall"
                    else room_num_or_wall
                )
                for direction, room_num_or_wall in self.room_config_str[
                    self.names["room"][room_num]
                ].items()
            }

        self.object_init_config_str = {}
        self.object_init_config_str["static"] = {
            self.names["static_objects"][object_num]: {
                self.names["room"][room_num]: prob
                for room_num, prob in room_num_dist.items()
            }
            for object_num, room_num_dist in self.object_init_config["static"].items()
        }

        self.object_init_config_str["independent"] = {
            self.names["independent_objects"][object_num]: {
                self.names["room"][room_num]: prob
                for room_num, prob in nesws_dist.items()
            }
            for object_num, nesws_dist in self.object_init_config["independent"].items()
        }

        self.object_init_config_str["dependent"] = {
            self.names["dependent_objects"][object_num]: {
                self.names["room"][room_num]: prob
                for room_num, prob in room_num_dist.items()
            }
            for object_num, room_num_dist in self.object_init_config[
                "dependent"
            ].items()
        }

        self.object_init_config_str["agent"] = {
            "agent": {
                self.names["room"][room_num]: prob
                for room_num, prob in self.object_init_config["agent"][0].items()
            }
        }

        for object_type in ["static", "independent", "dependent", "agent"]:
            for _, room_num_dist in self.object_init_config_str[object_type].items():
                assert (
                    abs(sum(room_num_dist.values()) - 1) < EPSILON
                ), f"{sum(room_num_dist.values())} should be close to 1"

        self.object_transition_config_str = {}

        self.object_transition_config_str["static"] = {
            self.names["static_objects"][object_num]: None
            for object_num in self.object_transition_config["static"]
        }

        self.object_transition_config_str["independent"] = {
            self.names["independent_objects"][object_num]: {
                self.names["room"][room_num]: dist
                for room_num, dist in room_num_dist.items()
            }
            for object_num, room_num_dist in self.object_transition_config[
                "independent"
            ].items()
        }

        self.object_transition_config_str["dependent"] = {
            self.names["dependent_objects"][object_num]: {
                self.names["independent_objects"][human_num]: prob
                for human_num, prob in ind_dist.items()
            }
            for object_num, ind_dist in self.object_transition_config[
                "dependent"
            ].items()
        }
        self.object_transition_config_str["agent"] = self.object_transition_config[
            "agent"
        ]

    def _create_question_probs(self) -> None:
        """Create a question probability of an object being asked.

        This is applied in a global manner.

        """
        self.question_probs = {}
        num_objects_except_agent = (
            self.num_static_objects
            + self.num_independent_objects
            + self.num_dependent_objects
        )

        for object_type in ["static", "independent", "dependent"]:
            if self.same_question_probs:
                self.question_probs[object_type] = {
                    object_name: 1 / num_objects_except_agent
                    for object_name in self.names[f"{object_type}_objects"]
                }
            else:
                self.question_probs[object_type] = {
                    object_name: np.random.beta(0.5, 0.5, 1).item()
                    for object_name in self.names[f"{object_type}_objects"]
                }

        # we don't want to ask a question about the agent.
        self.question_probs["agent"] = {"agent": 0}

        denominator = 0
        for object_type, object_probs in self.question_probs.items():
            denominator += sum(object_probs.values())

        if denominator == 0:
            raise ValueError("There should be at least one object to ask a question.")

        for object_type, object_probs in self.question_probs.items():
            for object_name, prob in object_probs.items():
                self.question_probs[object_type][object_name] = prob / denominator

        all_probs = [
            prob
            for object_probs in self.question_probs.values()
            for prob in object_probs.values()
        ]

        assert (
            abs(sum(all_probs) - 1) < EPSILON
        ), f"{sum(all_probs)} should be close to 1"

    def _visualize_grids(
        self, figsize: tuple[int, int] = (15, 15), cell_text_size: int = 10
    ) -> None:
        plt.figure(figsize=figsize)
        num_rows = len(self.grid)
        num_cols = len(self.grid[0])

        # Create the first subplot for the "Before" plot
        plt.subplot(121)
        plt.title("Before removing the disconnected rooms")

        for row in range(num_rows):
            for col in range(num_cols):
                cell_content = self.previous_grid[row][col]
                if cell_content != 0:
                    color = "white"
                else:
                    color = "black"
                plt.gca().add_patch(
                    plt.Rectangle((col, num_rows - 1 - row), 1, 1, facecolor=color)
                )

        plt.gca().set_aspect("equal")
        plt.gca().set_xticks(range(num_cols + 1))
        plt.gca().set_yticks(range(num_rows + 1))
        plt.gca().grid(which="both")

        # Create the second subplot for the "After" plot
        plt.subplot(122)
        plt.title("After removing the disconnected rooms")

        for row in range(num_rows):
            for col in range(num_cols):
                cell_content = self.grid[row][col]
                if cell_content != 0:
                    color = "white"
                    room_index = self.room_indexes.index((row, col))
                    text = "\n".join(self.names["room"][room_index].split("_"))

                else:
                    color = "black"
                    text = ""
                plt.gca().add_patch(
                    plt.Rectangle((col, num_rows - 1 - row), 1, 1, facecolor=color)
                )
                plt.text(
                    col + 0.5,
                    num_rows - 1 - row + 0.5,
                    text,
                    ha="center",
                    va="center",
                    color="black",
                    fontsize=cell_text_size,
                )
        plt.gca().set_aspect("equal")
        plt.gca().set_xticks(range(num_cols + 1))
        plt.gca().set_yticks(range(num_rows + 1))
        plt.gca().grid(which="both")

        plt.show()
